"""
Queue backend abstraction manager.
"""

import collections
import traceback

import tornado.web

from ..interface.models.rest_models import rest_model
from ..interface.models.task_models import PriorityEnum, TaskStatusEnum
from ..interface.models.records import RecordStatusEnum
from ..interface.models.model_builder import build_procedure
from ..procedures import check_procedure_available, get_procedure_parser
from ..services import initialize_service
from ..web_handlers import APIHandler


class TaskQueueHandler(APIHandler):
    """
    Handles task management (querying/adding/modifying tasks)
    """

    _required_auth = "compute"

    def post(self):
        """Posts new tasks to the task queue."""

        body_model, response_model = rest_model("task_queue", "post")
        body = self.parse_bodymodel(body_model)

        # Format and submit tasks
        if not check_procedure_available(body.meta.procedure):
            raise tornado.web.HTTPError(status_code=400, reason="Unknown procedure {}.".format(body.meta.procedure))

        procedure_parser = get_procedure_parser(body.meta.procedure, self.storage, self.logger)

        # Verify the procedure
        verify = procedure_parser.verify_input(body)
        if verify is not True:
            raise tornado.web.HTTPError(status_code=400, reason=verify)

        payload = procedure_parser.submit_tasks(body)
        response = response_model(**payload)

        self.logger.info("POST: TaskQueue -  Added {} tasks.".format(response.meta.n_inserted))
        self.write(response)

    def get(self):
        """Gets task information from the task queue"""

        body_model, response_model = rest_model("task_queue", "get")
        body = self.parse_bodymodel(body_model)

        tasks = self.storage.get_queue(**{**body.data.dict(), **body.meta.dict()})
        response = response_model(**tasks)

        self.logger.info("GET: TaskQueue - {} pulls.".format(len(response.data)))
        self.write(response)

    def put(self):
        """Modifies tasks in the task queue"""

        body_model, response_model = rest_model("task_queue", "put")
        body = self.parse_bodymodel(body_model)

        if (body.data.id is None) and (body.data.base_result is None):
            raise tornado.web.HTTPError(status_code=400, reason="Id or ResultId must be specified.")

        if body.meta.operation == "restart":
            d = body.data.dict()
            d.pop("new_tag", None)
            d.pop("new_priority", None)
            tasks_updated = self.storage.queue_reset_status(**d, reset_error=True)
            data = {"n_updated": tasks_updated}
        elif body.meta.operation == "regenerate":
            tasks_updated = 0
            result_data = self.storage.get_procedures(id=body.data.base_result)["data"]

            new_tag = body.data.new_tag
            if body.data.new_priority is None:
                new_priority = PriorityEnum.NORMAL
            else:
                new_priority = PriorityEnum(int(body.data.new_priority))

            for r in result_data:
                model = build_procedure(r)

                # Only regenerate the task if the base record is not complete
                # This will not do anything if the task already exists
                if model.status != RecordStatusEnum.complete:
                    procedure_parser = get_procedure_parser(model.procedure, self.storage, self.logger)

                    task_info = procedure_parser.create_tasks([model], tag=new_tag, priority=new_priority)
                    n_inserted = task_info["meta"]["n_inserted"]
                    tasks_updated += n_inserted

                    # If we inserted a new task, then also reset base result statuses
                    # (ie, if it was running, then it obviously isn't since we made a new task)
                    if n_inserted > 0:
                        self.storage.reset_base_result_status(id=body.data.base_result)

                data = {"n_updated": tasks_updated}
        elif body.meta.operation == "modify":
            tasks_updated = self.storage.queue_modify_tasks(
                id=body.data.id,
                base_result=body.data.base_result,
                new_tag=body.data.new_tag,
                new_priority=body.data.new_priority,
            )
            data = {"n_updated": tasks_updated}
        else:
            raise tornado.web.HTTPError(status_code=400, reason=f"Operation '{operation}' is not valid.")

        response = response_model(data=data, meta={"errors": [], "success": True, "error_description": False})

        self.logger.info(f"PUT: TaskQueue - Operation: {body.meta.operation} - {tasks_updated}.")
        self.write(response)


class ServiceQueueHandler(APIHandler):
    """
    Handles service management (querying/add/modifying)
    """

    _required_auth = "compute"

    def post(self):
        """Posts new services to the service queue."""

        body_model, response_model = rest_model("service_queue", "post")
        body = self.parse_bodymodel(body_model)

        new_services = []
        for service_input in body.data:
            # Get molecules with ids
            if isinstance(service_input.initial_molecule, list):
                molecules = self.storage.get_add_molecules_mixed(service_input.initial_molecule)["data"]
                if len(molecules) != len(service_input.initial_molecule):
                    raise KeyError("We should catch this error.")
            else:
                molecules = self.storage.get_add_molecules_mixed([service_input.initial_molecule])["data"][0]

            # Update the input and build a service object
            service_input = service_input.copy(update={"initial_molecule": molecules})
            new_services.append(
                initialize_service(
                    self.storage, self.logger, service_input, tag=body.meta.tag, priority=body.meta.priority
                )
            )

        ret = self.storage.add_services(new_services)
        ret["data"] = {"ids": ret["data"], "existing": ret["meta"]["duplicates"]}
        ret["data"]["submitted"] = list(set(ret["data"]["ids"]) - set(ret["meta"]["duplicates"]))
        response = response_model(**ret)

        self.logger.info("POST: ServiceQueue -  Added {} services.\n".format(response.meta.n_inserted))
        self.write(response)

    def get(self):
        """Gets information about services from the service queue."""

        body_model, response_model = rest_model("service_queue", "get")
        body = self.parse_bodymodel(body_model)

        ret = self.storage.get_services(**{**body.data.dict(), **body.meta.dict()})
        response = response_model(**ret)

        self.logger.info("GET: ServiceQueue - {} pulls.\n".format(len(response.data)))
        self.write(response)

    def put(self):
        """Modifies services in the service queue"""

        body_model, response_model = rest_model("service_queue", "put")
        body = self.parse_bodymodel(body_model)

        if (body.data.id is None) and (body.data.procedure_id is None):
            raise tornado.web.HTTPError(status_code=400, reason="Id or ProcedureId must be specified.")

        if body.meta.operation == "restart":
            updates = self.storage.update_service_status("running", **body.data.dict())
            data = {"n_updated": updates}
        else:
            raise tornado.web.HTTPError(status_code=400, reason=f"Operation '{operation}' is not valid.")

        response = response_model(data=data, meta={"errors": [], "success": True, "error_description": False})

        self.logger.info(f"PUT: TaskQueue - Operation: {body.meta.operation} - {updates}.")
        self.write(response)


class QueueManagerHandler(APIHandler):
    """
    Manages the task queue.

    Used by compute managers for getting tasks, posting completed tasks, etc.
    """

    _required_auth = "queue"

    @staticmethod
    def _get_name_from_metadata(meta):
        """
        Form the canonical name string.
        """
        ret = meta.cluster + "-" + meta.hostname + "-" + meta.uuid
        return ret

    @staticmethod
    def insert_complete_tasks(storage_socket, body, logger):

        results = body.data
        meta = body.meta
        task_ids = list(results.keys())

        manager_name = QueueManagerHandler._get_name_from_metadata(meta)
        logger.info("QueueManager: Received completed tasks from {}.".format(manager_name))
        logger.info("              Task ids: " + " ".join(task_ids))

        # Pivot data so that we group all results in categories
        new_results = collections.defaultdict(list)

        queue = storage_socket.get_queue(id=task_ids)["data"]
        queue = {v.id: v for v in queue}

        error_data = []

        task_success = 0
        task_failures = 0
        task_totals = len(results.items())

        for task_id, result in results.items():
            try:
                #################################################################
                # Perform some checks for consistency
                #################################################################
                existing_task_data = queue.get(task_id, None)

                # For the first three checks, don't add an error to error_data
                # We don't want to modify the queue in these cases

                # Does the task exist?
                if existing_task_data is None:
                    logger.warning(f"Task id {task_id} does not exist in the task queue.")
                    task_failures += 1

                # Is the task in the running state
                elif existing_task_data.status != TaskStatusEnum.running:
                    logger.warning(f"Task id {task_id} is not in the running state.")
                    task_failures += 1

                # Was the manager that sent the data the one that was assigned?
                elif existing_task_data.manager != manager_name:
                    logger.warning(f"Task id {task_id} belongs to {existing_task_data.manager}, not this manager")
                    task_failures += 1

                # Failed task
                elif result["success"] is False:
                    if "error" not in result:
                        error = {"error_type": "not_supplied", "error_message": "No error message found on task."}
                    else:
                        error = result["error"]

                    logger.debug(
                        "Task id {key} did not complete successfully:\n"
                        "error_type: {error_type}\nerror_message: {error_message}".format(key=str(task_id), **error)
                    )

                    error_data.append((task_id, error))
                    task_failures += 1

                # Success!
                else:
                    parser = queue[task_id].parser
                    new_results[parser].append(
                        {"result": result, "task_id": task_id, "base_result": queue[task_id].base_result}
                    )
                    task_success += 1

            except Exception:
                msg = "Internal FractalServer Error:\n" + traceback.format_exc()
                error = {"error_type": "internal_fractal_error", "error_message": msg}
                logger.error("update: ERROR\n{}".format(msg))
                error_data.append((task_id, error))
                task_failures += 1

        if task_totals:
            logger.info(
                "QueueManager: Found {} complete tasks ({} successful, {} failed).".format(
                    task_totals, task_success, task_failures
                )
            )

        # Run output parsers and handle completed tasks
        completed = []
        for k, v in new_results.items():
            procedure_parser = get_procedure_parser(k, storage_socket, logger)
            com = procedure_parser.handle_completed_output(v)
            completed.extend(com)

        storage_socket.queue_mark_error(error_data)
        return len(completed), len(error_data)

    def get(self):
        """Pulls new tasks from the task queue"""

        body_model, response_model = rest_model("queue_manager", "get")
        body = self.parse_bodymodel(body_model)

        # Figure out metadata and kwargs
        name = self._get_name_from_metadata(body.meta)

        # Grab new tasks and write out
        new_tasks = self.storage.queue_get_next(
            name, body.meta.programs, body.meta.procedures, limit=body.data.limit, tag=body.meta.tag
        )
        response = response_model(
            **{
                "meta": {
                    "n_found": len(new_tasks),
                    "success": True,
                    "errors": [],
                    "error_description": "",
                    "missing": [],
                },
                "data": new_tasks,
            }
        )
        self.write(response)

        self.logger.info("QueueManager: Served {} tasks.".format(response.meta.n_found))

        # Update manager logs
        self.storage.manager_update(name, submitted=len(new_tasks), **body.meta.dict())

    def post(self):
        """Posts complete tasks to the task queue"""

        body_model, response_model = rest_model("queue_manager", "post")
        body = self.parse_bodymodel(body_model)

        success, error = self.insert_complete_tasks(self.storage, body, self.logger)

        completed = success + error

        response = response_model(
            **{
                "meta": {
                    "n_inserted": completed,
                    "duplicates": [],
                    "validation_errors": [],
                    "success": True,
                    "errors": [],
                    "error_description": "",
                },
                "data": True,
            }
        )
        self.write(response)
        self.logger.info("QueueManager: Inserted {} complete tasks.".format(len(body.data)))

        # Update manager logs
        name = self._get_name_from_metadata(body.meta)
        self.storage.manager_update(name, completed=completed, failures=error)

    def put(self):
        """
        Various manager manipulation operations
        """

        ret = True

        body_model, response_model = rest_model("queue_manager", "put")
        body = self.parse_bodymodel(body_model)

        name = self._get_name_from_metadata(body.meta)
        op = body.data.operation
        if op == "startup":
            self.storage.manager_update(
                name, status="ACTIVE", configuration=body.data.configuration, **body.meta.dict(), log=True
            )
            self.logger.info("QueueManager: New active manager {} detected.".format(name))

        elif op == "shutdown":
            nshutdown = self.storage.queue_reset_status(manager=name, reset_running=True)
            self.storage.manager_update(name, returned=nshutdown, status="INACTIVE", **body.meta.dict(), log=True)

            self.logger.info(
                "QueueManager: Shutdown of manager {} detected, recycling {} incomplete tasks.".format(name, nshutdown)
            )

            ret = {"nshutdown": nshutdown}

        elif op == "heartbeat":
            self.storage.manager_update(name, status="ACTIVE", **body.meta.dict(), log=True)
            self.logger.debug("QueueManager: Heartbeat of manager {} detected.".format(name))

        else:
            msg = "Operation '{}' not understood.".format(op)
            raise tornado.web.HTTPError(status_code=400, reason=msg)

        response = response_model(**{"meta": {}, "data": ret})
        self.write(response)

        # Update manager logs
        # TODO: ????


class ComputeManagerHandler(APIHandler):
    """
    Handles management/status querying of managers
    """

    _required_auth = "admin"

    def get(self):
        """Gets manager information from the task queue"""

        body_model, response_model = rest_model("manager", "get")
        body = self.parse_bodymodel(body_model)

        self.logger.info("GET: ComputeManagerHandler")
        managers = self.storage.get_managers(**{**body.data.dict(), **body.meta.dict()})

        # remove passwords?
        # TODO: Are passwords stored anywhere else? Other kinds of passwords?
        for m in managers["data"]:
            if "configuration" in m and isinstance(m["configuration"], dict) and "server" in m["configuration"]:
                m["configuration"]["server"].pop("password", None)

        response = response_model(**managers)
        self.write(response)



# ORM Base
# Collections ORMs
from .collections_models import CollectionORM, DatasetORM, ReactionDatasetORM

# Results and procedures ORMs
from .results_models import (
    BaseResultORM,
    GridOptimizationProcedureORM,
    OptimizationHistory,
    OptimizationProcedureORM,
    ResultORM,
    TorsionDriveProcedureORM,
    Trajectory,
    WavefunctionStoreORM,
)
from .sql_base import Base, MsgpackExt

# ORM general models
from .sql_models import (
    AccessLogORM,
    KeywordsORM,
    KVStoreORM,
    MoleculeORM,
    QueueManagerLogORM,
    QueueManagerORM,
    ServerStatsLogORM,
    ServiceQueueORM,
    TaskQueueORM,
    UserORM,
    VersionsORM,
)



"""id(pirmary key) for torsion_init_mol

Revision ID: 1604623c481a
Revises: fb5bd88ae2f3
Create Date: 2020-07-02 18:42:17.267792

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "1604623c481a"
down_revision = "fb5bd88ae2f3"
branch_labels = None
depends_on = None


def upgrade():

    # Removes (harmless) duplicate rows
    op.execute(
        "DELETE FROM torsion_init_mol_association a USING \
    (SELECT MIN(ctid) as ctid, torsion_id, molecule_id \
        FROM torsion_init_mol_association \
        GROUP BY torsion_id, molecule_id HAVING COUNT(*) > 1 \
      ) b \
      WHERE a.torsion_id = b.torsion_id and a.molecule_id = b.molecule_id \
      AND a.ctid <> b.ctid"
    )

    op.execute("alter table torsion_init_mol_association add primary key (torsion_id, molecule_id)")


def downgrade():
    op.execute("alter table torsion_init_mol_association drop constraint torsion_init_mol_association_pkey")



"""
Optimization procedure/task
"""

from typing import List, Optional

import qcelemental as qcel
import qcengine as qcng

from .base import BaseTasks
from ..interface.models import Molecule, OptimizationRecord, QCSpecification, ResultRecord, TaskRecord, KeywordSet
from ..interface.models.task_models import PriorityEnum
from .procedures_util import parse_single_tasks, form_qcinputspec_schema


class OptimizationTasks(BaseTasks):
    """
    Optimization task manipulation
    """

    def verify_input(self, data):
        program = data.meta.program.lower()
        if program not in qcng.list_all_procedures():
            return "Procedure '{}' not available in QCEngine.".format(program)

        program = data.meta.qc_spec["program"].lower()
        if program not in qcng.list_all_programs():
            return "Program '{}' not available in QCEngine.".format(program)

        return True

    def parse_input(self, data, duplicate_id="hash_index"):
        """Parse input json into internally appropriate format

        json_data = {
            "meta": {
                "procedure": "optimization",
                "option": "default",
                "program": "geometric",
                "qc_meta": {
                    "driver": "energy",
                    "method": "HF",
                    "basis": "sto-3g",
                    "keywords": "default",
                    "program": "psi4"
                },
            },
            "data": ["mol_id_1", "mol_id_2", ...],
        }

        qc_schema_input = {
            "molecule": {
                "geometry": [
                    0.0,  0.0, -0.6,
                    0.0,  0.0,  0.6,
                ],
                "symbols": ["H", "H"],
                "connectivity": [[0, 1, 1]]
            },
            "driver": "gradient",
            "model": {
                "method": "HF",
                "basis": "sto-3g"
            },
            "keywords": {},
        }
        json_data = {
            "keywords": {
                "coordsys": "tric",
                "maxiter": 100,
                "program": "psi4"
            },
        }

        """

        # Get the optimization specification from the input meta dictionary
        opt_spec = data.meta

        # We should only have gotten here if procedure is 'optimization'
        assert opt_spec.procedure.lower() == "optimization"

        # Grab the tag and priority if available
        tag = opt_spec.tag
        priority = opt_spec.priority

        # Handle (optimization) keywords, which may be None
        # TODO: These are not stored in the keywords table (yet)
        opt_keywords = {} if opt_spec.keywords is None else opt_spec.keywords

        # Set the program used for gradient evaluations. This is stored in the input qcspec
        # but the QCInputSpecification does not have a place for program. So instead
        # we move it to the optimization keywords
        opt_keywords["program"] = opt_spec.qc_spec["program"]

        # Pull out the QCSpecification from the input
        qc_spec_dict = data.meta.qc_spec

        # Handle qc specification keywords, which may be None
        qc_keywords = qc_spec_dict.get("keywords", None)
        if qc_keywords is not None:
            # The keywords passed in may contain the entire KeywordSet.
            # But the QCSpec will only hold the ID
            qc_keywords = self.storage.get_add_keywords_mixed([qc_keywords])["data"][0]
            if qc_keywords is None:
                raise KeyError("Could not find requested KeywordsSet from id key.")
            qc_spec_dict["keywords"] = qc_keywords.id

        # Now that keywords are fixed we can do this
        qc_spec = QCSpecification(**qc_spec_dict)

        # Add all the initial molecules to the database
        # TODO: WARNING WARNING if get_add_molecules_mixed is modified to handle duplicates
        #       correctly, you must change some pieces later in this function
        molecule_list = self.storage.get_add_molecules_mixed(data.data)["data"]

        # Keep molecule IDs that are not None
        # Molecule IDs may be None if they are duplicates (ie, the same molecule was listed twice
        # in data.data) or an id specified in data.data was invalid
        valid_molecule_idx = [idx for idx, mol in enumerate(molecule_list) if mol is not None]
        valid_molecules = [x for x in molecule_list if x is not None]

        # Create all OptimizationRecords
        all_opt_records = []
        for mol in valid_molecules:
            # TODO fix handling of protocols (perhaps after hardening rest models)
            opt_data = {
                "initial_molecule": mol.id,
                "qc_spec": qc_spec,
                "keywords": opt_keywords,
                "program": opt_spec.program,
            }
            if hasattr(opt_spec, "protocols"):
                opt_data["protocols"] = data.meta.protocols

            opt_rec = OptimizationRecord(**opt_data)
            all_opt_records.append(opt_rec)

        # Add all the procedures in a single function call
        # NOTE: Because get_add_molecules_mixed returns None for duplicate
        # molecules (or when specifying incorrect ids),
        # all_opt_records should never contain duplicates
        ret = self.storage.add_procedures(all_opt_records)

        # Get all procedure IDs (may be new or existing)
        # These will be in the order we sent to add_results
        all_opt_ids = ret["data"]
        existing_ids = ret["meta"]["duplicates"]

        # Assing ids to the optimization records
        for idx in range(len(all_opt_records)):
            r = all_opt_records[idx].copy(update={"id": all_opt_ids[idx]})
            all_opt_records[idx] = r

        # Now generate all the tasks, but only for results that don't exist already
        self.create_tasks(
            all_opt_records, valid_molecules, [qc_keywords] * len(all_opt_records), tag=tag, priority=priority
        )

        # Keep the returned result id list in the same order as the input molecule list
        # If a molecule was None, then the corresponding result ID will be None
        # (since the entry in valid_molecule_idx will be missing). Ditto for molecules specified
        # more than once in the argument to this function
        opt_ids = [None] * len(molecule_list)
        for idx, result_id in zip(valid_molecule_idx, all_opt_ids):
            opt_ids[idx] = result_id

        return opt_ids, existing_ids

    def create_tasks(
        self,
        records: List[OptimizationRecord],
        molecules: Optional[List[Molecule]] = None,
        qc_keywords: Optional[List[KeywordSet]] = None,
        tag: Optional[str] = None,
        priority: Optional[PriorityEnum] = None,
    ):

        # Find the molecule keywords specified in the records
        rec_mol_ids = [x.initial_molecule for x in records]

        # If not specified when calling this function, load them from the database
        # TODO: there can be issues with duplicate molecules. So we have to go one by one
        if molecules is None:
            molecules = [self.storage.get_molecules(x)["data"][0] for x in rec_mol_ids]

        # Check id to make sure the molecules match the ids in the records
        mol_ids = [x.id for x in molecules]
        if rec_mol_ids != mol_ids:
            raise ValueError(f"Given molecule ids {str(mol_ids)} do not match those in records: {str(rec_mol_ids)}")

        # Do the same as above but with with qc specification keywords
        rec_qc_kw_ids = [x.qc_spec.keywords for x in records]
        if qc_keywords is None:
            qc_keywords = [self.storage.get_keywords(x)["data"][0] if x is not None else None for x in rec_qc_kw_ids]

        qc_kw_ids = [x.id if x is not None else None for x in qc_keywords]
        if rec_qc_kw_ids != qc_kw_ids:
            raise ValueError(f"Given keyword ids {str(qc_kw_ids)} do not match those in records: {str(rec_qc_kw_ids)}")

        new_tasks = []
        for rec, mol, kw in zip(records, molecules, qc_keywords):
            inp = self._build_schema_input(rec, mol, kw)

            # Build task object
            task = TaskRecord(
                **{
                    "spec": {
                        "function": "qcengine.compute_procedure",
                        "args": [inp.dict(), rec.program],
                        "kwargs": {},
                    },
                    "parser": "optimization",
                    # TODO This is pretty whacked. Fix column names at some point
                    "program": rec.qc_spec.program,
                    "procedure": rec.program,
                    "tag": tag,
                    "priority": priority,
                    "base_result": rec.id,
                }
            )

            new_tasks.append(task)

        return self.storage.queue_submit(new_tasks)

    def handle_completed_output(self, opt_outputs):
        """Save the results of the procedure.
        It must make sure to save the results in the results table
        including the task_id in the TaskQueue table
        """

        completed_tasks = []
        updates = []
        for output in opt_outputs:
            rec = self.storage.get_procedures(id=output["base_result"])["data"][0]
            rec = OptimizationRecord(**rec)

            procedure = output["result"]

            # Adds the results to the database and sets the ids inside the dictionary
            self.retrieve_outputs(procedure)

            # Add initial and final molecules
            update_dict = {}
            update_dict["stdout"] = procedure.get("stdout", None)
            update_dict["stderr"] = procedure.get("stderr", None)
            update_dict["error"] = procedure.get("error", None)

            initial_mol, final_mol = self.storage.add_molecules(
                [Molecule(**procedure["initial_molecule"]), Molecule(**procedure["final_molecule"])]
            )["data"]
            assert initial_mol == rec.initial_molecule
            update_dict["final_molecule"] = final_mol

            # Parse trajectory computations and add task_id
            traj_dict = {k: v for k, v in enumerate(procedure["trajectory"])}

            # Add results for the trajectory to the database
            for k, v in traj_dict.items():
                self.retrieve_outputs(v)

            results = parse_single_tasks(self.storage, traj_dict, rec.qc_spec)
            for k, v in results.items():
                results[k] = ResultRecord(**v)

            ret = self.storage.add_results(list(results.values()))
            update_dict["trajectory"] = ret["data"]
            update_dict["energies"] = procedure["energies"]
            update_dict["provenance"] = procedure["provenance"]

            rec = OptimizationRecord(**{**rec.dict(), **update_dict})
            updates.append(rec)
            completed_tasks.append(output["task_id"])

        self.storage.update_procedures(updates)
        self.storage.queue_mark_complete(completed_tasks)

        return completed_tasks

    @staticmethod
    def _build_schema_input(
        record: OptimizationRecord, initial_molecule: "Molecule", qc_keywords: Optional["KeywordSet"] = None
    ) -> "OptimizationInput":
        """
        Creates a OptimizationInput schema.
        """

        assert record.initial_molecule == initial_molecule.id
        if record.qc_spec.keywords:
            assert record.qc_spec.keywords == qc_keywords.id

        qcinput_spec = form_qcinputspec_schema(record.qc_spec, keywords=qc_keywords)

        model = qcel.models.OptimizationInput(
            id=record.id,
            initial_molecule=initial_molecule,
            keywords=record.keywords,
            extras=record.extras,
            hash_index=record.hash_index,
            input_specification=qcinput_spec,
            protocols=record.protocols,
        )
        return model



"""
Procedure for a single computational task (energy, gradient, etc)
"""

from typing import List, Optional

import qcelemental as qcel
import qcengine as qcng

from .base import BaseTasks
from ..interface.models import Molecule, ResultRecord, TaskRecord, KeywordSet
from ..interface.models.task_models import PriorityEnum

_wfn_return_names = set(qcel.models.results.WavefunctionProperties._return_results_names)
_wfn_all_fields = set(qcel.models.results.WavefunctionProperties.__fields__.keys())


class SingleResultTasks(BaseTasks):
    """A task generator for a single QC computation task.

    This is a single quantum calculation, unique by program, driver, method, basis, keywords, molecule.
    """

    def verify_input(self, data):
        program = data.meta.program.lower()
        if program not in qcng.list_all_programs():
            return f"Program '{program}' not available in QCEngine."

        if data.meta.dict().get("protocols", None) is not None:
            try:
                qcel.models.results.ResultProtocols(**data.meta.protocols)
            except Exception as e:
                return f"Could not validate protocols: {str(e)}"

        return True

    def parse_input(self, data):
        """Parse input json into internally appropriate format


        Format of the input data:
        data = {
            "meta": {
                "procedure": "single",
                "driver": "energy",
                "method": "HF",
                "basis": "sto-3g",
                "keywords": "default",
                "program": "psi4"
                },
            },
            "data": ["mol_id_1", "mol_id_2", ...],
        }

        """

        # Get the qcspec from the input metadata dictionary
        qc_spec_dict = data.meta.dict()

        # We should only have gotten here if procedure is 'single'
        procedure = qc_spec_dict.pop("procedure")
        assert procedure.lower() == "single"

        # Grab the tag and priority if available
        # These are not used in the ResultRecord, so we can pop them
        tag = qc_spec_dict.pop("tag")
        priority = qc_spec_dict.pop("priority")

        # Handle keywords, which may be None
        if data.meta.keywords is not None:
            # QCSpec will only hold the ID
            keywords = self.storage.get_add_keywords_mixed([data.meta.keywords])["data"][0]
            if keywords is None:
                raise KeyError("Could not find requested KeywordsSet from id key.")
            qc_spec_dict["keywords"] = keywords.id
        else:
            keywords = None
            qc_spec_dict["keywords"] = None

        # Will also validate the qc_spec
        # TODO - do this after migrating results to QCSpecification?
        # qc_spec = QCSpecification(**qc_spec_dict)

        # Add all the molecules to the database
        # TODO: WARNING WARNING if get_add_molecules_mixed is modified to handle duplicates
        #       correctly, you must change some pieces later in this function
        input_molecules = data.data
        molecule_list = self.storage.get_add_molecules_mixed(input_molecules)["data"]

        # Keep molecule IDs that are not None
        # Molecule IDs may be None if they are duplicates (ie, the same molecule was listed twice
        # in data.data) or an id specified in data.data was invalid
        valid_molecule_idx = [idx for idx, mol in enumerate(molecule_list) if mol is not None]
        valid_molecules = [x for x in molecule_list if x is not None]

        # Create ResultRecords for everything
        all_result_records = []
        for mol in valid_molecules:
            record = ResultRecord(**qc_spec_dict.copy(), molecule=mol.id)
            all_result_records.append(record)

        # Add all results in a single function call
        # NOTE: Because get_add_molecules_mixed returns None for duplicate
        # molecules (or when specifying incorrect ids),
        # all_result_records should never contain duplicates
        ret = self.storage.add_results(all_result_records)

        # Get all the result IDs (may be new or existing)
        # These will be in the order we sent to add_results
        all_result_ids = ret["data"]
        existing_ids = ret["meta"]["duplicates"]

        # Assign ids to the result records
        for idx in range(len(all_result_records)):
            r = all_result_records[idx].copy(update={"id": all_result_ids[idx]})
            all_result_records[idx] = r

        # Now generate all the tasks, but only for results that don't exist already
        self.create_tasks(
            all_result_records, valid_molecules, [keywords] * len(all_result_records), tag=tag, priority=priority
        )

        # Keep the returned result id list in the same order as the input molecule list
        # If a molecule was None, then the corresponding result ID will be None
        # (since the entry in valid_molecule_idx will be missing). Ditto for molecules specified
        # more than once in the argument to this function
        results_ids = [None] * len(molecule_list)
        for idx, result_id in zip(valid_molecule_idx, all_result_ids):
            results_ids[idx] = result_id

        return results_ids, existing_ids

    def create_tasks(
        self,
        records: List[ResultRecord],
        molecules: Optional[List[Molecule]] = None,
        keywords: Optional[List[KeywordSet]] = None,
        tag: Optional[str] = None,
        priority: Optional[PriorityEnum] = None,
    ):
        """
        Creates TaskRecord objects based on a record and molecules/keywords

        If molecules are not given, then they are loaded from the database (with the id given in the result record)
        The same applies to the keywords.

        Parameters
        ----------
        records: List[ResultRecord]
            Records for which to create the TaskRecord object
        molecules: Optional[List[Molecule]]
            Molecules to be applied to the records. If given, must be the same length as records, and
            must be in the same order. They must have been added to the database already and have an id.
        keywords: Optional[KeywordSet]
            QC Keywords to use in the calculation. If given, must be the same length as records, and
            be in the same order. They must have been added to the database already and have an id.

        Returns
        -------
        List[TaskRecord]
            TaskRecords that can be added to the database
        """

        # Find the molecule keywords specified in the records
        rec_mol_ids = [x.molecule for x in records]

        # If not specified when calling this function, load them from the database
        # TODO: there can be issues with duplicate molecules. So we have to go one by one
        if molecules is None:
            molecules = [self.storage.get_molecules(x)["data"][0] for x in rec_mol_ids]

        # Check id to make sure the molecules match the ids in the records
        mol_ids = [x.id for x in molecules]
        if rec_mol_ids != mol_ids:
            raise ValueError(f"Given molecule ids {str(mol_ids)} do not match those in records: {str(rec_mol_ids)}")

        # Do the same as above but with with keywords
        rec_kw_ids = [x.keywords for x in records]
        if keywords is None:
            keywords = [self.storage.get_keywords(x)["data"][0] if x is not None else None for x in rec_kw_ids]

        kw_ids = [x.id if x is not None else None for x in keywords]
        if rec_kw_ids != kw_ids:
            raise ValueError(f"Given keyword ids {str(kw_ids)} do not match those in records: {str(rec_kw_ids)}")

        # Create QCSchema inputs and tasks for everything, too
        new_tasks = []

        for rec, mol, kw in zip(records, molecules, keywords):
            inp = self._build_schema_input(rec, mol, kw)
            inp.extras["_qcfractal_tags"] = {"program": rec.program, "keywords": rec.keywords}

            # Build task object
            task = TaskRecord(
                **{
                    "spec": {
                        "function": "qcengine.compute",  # todo: add defaults in models
                        "args": [inp.dict(), rec.program],
                        "kwargs": {},  # todo: add defaults in models
                    },
                    "parser": "single",
                    "program": rec.program,
                    "tag": tag,
                    "priority": priority,
                    "base_result": rec.id,
                }
            )

            new_tasks.append(task)

        return self.storage.queue_submit(new_tasks)

    def handle_completed_output(self, result_outputs):

        completed_tasks = []
        updates = []

        for output in result_outputs:
            # Find the existing result information in the database
            base_id = output["base_result"]
            existing_result = self.storage.get_results(id=base_id)
            if existing_result["meta"]["n_found"] != 1:
                raise KeyError(f"Could not find existing base result {base_id}")

            # Get the original result data from the dictionary
            existing_result = existing_result["data"][0]

            # Some consistency checks:
            # Is this marked as incomplete?
            # TODO: Check manager, although that information isn't sent to us right now
            if existing_result["status"] != "INCOMPLETE":
                self.logger.warning(f"Skipping returned results for base_id={base_id}, as it is not marked incomplete")
                continue

            rdata = output["result"]

            # Adds the results to the database and sets the appropriate fields
            # inside the dictionary
            self.retrieve_outputs(rdata)

            # Store Wavefunction data
            if rdata.get("wavefunction", False):
                wfn = rdata.get("wavefunction", False)
                available = set(wfn.keys()) - {"restricted", "basis"}
                return_map = {k: wfn[k] for k in wfn.keys() & _wfn_return_names}

                rdata["wavefunction"] = {
                    "available": list(available),
                    "restricted": wfn["restricted"],
                    "return_map": return_map,
                }

                # Extra fields are trimmed as we have a column *per* wavefunction structure.
                available_keys = wfn.keys() - _wfn_return_names
                if available_keys > _wfn_all_fields:
                    self.logger.warning(f"Too much wavefunction data for result {base_id}, removing extra data.")
                    available_keys &= _wfn_all_fields

                wavefunction_save = {k: wfn[k] for k in available_keys}
                wfn_data_id = self.storage.add_wavefunction_store([wavefunction_save])["data"][0]
                rdata["wavefunction_data_id"] = wfn_data_id

            # Create an updated ResultRecord based on the existing record and the new results
            # Double check to make sure everything is consistent
            assert existing_result["method"] == rdata["model"]["method"]
            assert existing_result["basis"] == rdata["model"]["basis"]
            assert existing_result["driver"] == rdata["driver"]
            assert existing_result["molecule"] == rdata["molecule"]["id"]

            # Result specific
            existing_result["extras"] = rdata["extras"]
            existing_result["return_result"] = rdata["return_result"]
            existing_result["properties"] = rdata["properties"]

            # Wavefunction data
            existing_result["wavefunction"] = rdata.get("wavefunction", None)
            existing_result["wavefunction_data_id"] = rdata.get("wavefunction_data_id", None)

            # Standard blocks
            existing_result["provenance"] = rdata["provenance"]
            existing_result["error"] = rdata["error"]
            existing_result["stdout"] = rdata["stdout"]
            existing_result["stderr"] = rdata["stderr"]
            existing_result["status"] = "COMPLETE"

            result = ResultRecord(**existing_result)

            # Add to the list to be updated
            updates.append(result)
            completed_tasks.append(output["task_id"])

        self.storage.update_results(updates)
        self.storage.queue_mark_complete(completed_tasks)

        return completed_tasks

    @staticmethod
    def _build_schema_input(
        record: ResultRecord, molecule: "Molecule", keywords: Optional["KeywordSet"] = None
    ) -> "ResultInput":
        """
        Creates an input schema for a single calculation
        """

        # Check for programmer sanity. Since we are building this input
        # right after creating the ResultRecord, these should never fail.
        # But would be very hard to debug
        assert record.molecule == molecule.id
        if record.keywords:
            assert record.keywords == keywords.id

        # Now start creating the "model" parameter for ResultInput
        model = {"method": record.method}

        if record.basis:
            model["basis"] = record.basis

        if not record.keywords:
            keywords = {}
        else:
            keywords = keywords.values

        if not record.protocols:
            protocols = {}
        else:
            protocols = record.protocols

        return qcel.models.AtomicInput(
            id=record.id,
            driver=record.driver.name,
            model=model,
            molecule=molecule,
            keywords=keywords,
            extras=record.extras,
            protocols=protocols,
        )



import json

import pytest

from ..common_models import KeywordSet, Molecule
from ..gridoptimization import GridOptimizationRecord
from ..records import OptimizationRecord, ResultRecord
from ..torsiondrive import TorsionDriveRecord

## Molecule hashes


def test_molecule_water_canary_hash():

    water_dimer_minima = Molecule.from_data(
        """
    0 1
    O  -1.551007  -0.114520   0.000000
    H  -1.934259   0.762503   0.000000
    H  -0.599677   0.040712   0.000000
    --
    O   1.350625   0.111469   0.000000
    H   1.680398  -0.373741  -0.758561
    H   1.680398  -0.373741   0.758561
    """,
        dtype="psi4",
    )
    assert water_dimer_minima.get_hash() == "42f3ac52af52cf2105c252031334a2ad92aa911c"

    # Check orientation
    mol = water_dimer_minima.orient_molecule()
    assert mol.get_hash() == "632490a0601500bfc677e9277275f82fbc45affe"

    frag_0 = mol.get_fragment(0, orient=True)
    frag_1 = mol.get_fragment(1, orient=True)
    assert frag_0.get_hash() == "d0b499739f763e8d3a5556b4ddaeded6a148e4d5"
    assert frag_1.get_hash() == "bdc1f75bd1b7b999ff24783d7c1673452b91beb9"


@pytest.mark.parametrize(
    "geom",
    [[0, 0, 0, 0, 5, 0], [0, 0, 0, 0, 5, 0 + 1.0e-12], [0, 0, 0, 0, 5, 0 - 1.0e-12], [0, 0, 0, 0, 5, 0 + 1.0e-7]],
)  # yapf: disable
def test_molecule_geometry_canary_hash(geom):

    mol = Molecule(geometry=geom, symbols=["H", "H"])

    assert mol.get_hash() == "fb69e6744407b220a96d6ddab4ec2099619db791"


## Keyword Set hash


@pytest.mark.parametrize(
    "data, hash_index",
    [
        # Simple checks
        ({"values": {"hi": 5}}, "53042da4ac1af059816008631e4589a0cd0c98c6"),
        ({"values": {"HI": 5}}, "53042da4ac1af059816008631e4589a0cd0c98c6"),
        # Check options combinations
        ({"values": {"HI": [1, 2, 3], "All": [1.0e-5, 2.0e-3, 1.0e-16]}}, "888837dec981e2f9a1ef2ef2d33db6d239df65da"),
        (
            {"values": {"HI": [1, 2, 3], "All": [1.0e-5, 2.0e-3, 1.0e-16]}, "lowercase": False},
            "08f6f37fec721b753096d71f7dbbccb734e8a696",
        ),
        (
            {"values": {"HI": [1, 2, 3], "All": [1.0e-5, 2.0e-3, 1.0e-16]}, "exact_floats": True},
            "691c86202af20868bba457514e63c83d0444142e",
        ),
        (
            {"values": {"HI": [1, 2, 3], "All": [1.0e-5, 2.0e-3, 1.0e-16]}, "exact_floats": True, "lowercase": False},
            "a3a6eea9edec3c2901a76fa8a41d4f9db9b3405f",
        ),
        # Check recursive
        ({"values": {"d1": {"D2": [1.0e-17, 5]}}}, "55c6f1c7e610ce379feaa8a97854ea61974e0d92"),
        (
            {"values": {"d1": {"D2": [1.0e-17, 5]}}, "exact_floats": True, "lowercase": False},
            "639d207f6a58ed2974f9737c3240ebffdf459857",
        ),
        # Check hash_index build
        ({"values": {}, "hash_index": "waffles"}, "waffles"),
        ({"values": {}, "hash_index": "waffles", "build_index": True}, "bf21a9e8fbc5a3846fb05b4fa0859e0917b2202f"),
    ],
)  # yapf: disable
def test_keywords_canary_hash(data, hash_index):
    """
    Canary test aginst possible changes in the options hash_index
    """

    opt = KeywordSet(**data)

    assert hash_index == opt.hash_index, opt.values


@pytest.mark.parametrize(
    "data1, data2",
    [
        # Test case sensitivity
        ({"values": {}}, {"values": {}}),
        # Test float creation
        ({"values": {"CAPS": 5, "other": 4.0e-3}}, {"values": {"caps": 5, "other": 0.004}}),
        ({"values": {"other": 4.123e-5}}, {"values": {"OTHER": 0.00004123}}),
        # Test list of floats
        ({"values": {"other": [1.11e-2, 2.22e-3]}}, {"values": {"other": [0.0111, 0.00222]}}),
        # Test small floats
        ({"values": {"other": 1 + 1.0e-17}}, {"values": {"other": 1.0 - 1.0e-17}}),
        # Test dict order
        ({"values": {"a": 5, "b": 6, "c": None}}, {"values": {"b": 6, "a": 5, "c": None}}),
        # Check recusive
        (
            {"values": {"d1": {"D2": [0.0, 5], "d3": (3, 1.0e-17)}}},
            {"values": {"d1": {"d2": [1.0e-17, 5], "d3": (3, 0)}}},
        ),
    ],
)  # yapf: disable
def test_keywords_comparison_hash(data1, data2):
    """
    Ensure the hash_index finds collisions correctly before and after serialization.
    """
    opt1 = KeywordSet(**data1)
    opt2 = KeywordSet(**data2)

    # Check after serialization and rebuild index
    opt1s = KeywordSet(**json.loads(opt1.json()), build_index=True)
    opt2s = KeywordSet(**json.loads(opt2.json()), build_index=True)

    # Paranoid, try all combinations
    assert opt1.hash_index == opt2.hash_index
    assert opt1.hash_index == opt2s.hash_index
    assert opt1s.hash_index == opt2.hash_index
    assert opt1s.hash_index == opt2s.hash_index


@pytest.mark.parametrize("model", [ResultRecord, OptimizationRecord])
def test_hash_fields(model):
    assert "procedure" in model.get_hash_fields()
    assert "program" in model.get_hash_fields()


## Record hashes
_base_result = {
    "driver": "gradient",
    "method": "hf",
    "basis": "sto-3g",
    "keywords": None,
    "program": "prog",
    "molecule": "5c7896fb95d592ad07a2fe3b",
}


@pytest.mark.parametrize(
    "data, hash_index",
    [
        # Check same
        ({}, "e1e20d5c13c8ad7ba894f71af39f3c0884ef2aca"),
        ({"program": "PROG"}, "e1e20d5c13c8ad7ba894f71af39f3c0884ef2aca"),
        ({"method": "HF"}, "e1e20d5c13c8ad7ba894f71af39f3c0884ef2aca"),
        # All different
        ({"program": "prog2"}, "f8638510976e5146b6bef79995f6f14acd06f645"),
        ({"driver": "energy"}, "4c2c737694f9438c73b294d7f72249acbbd3b649"),
        ({"keywords": "5c7896fb95d592ad07a2fe3b"}, "32e115a991ffb85ac0c29bf4dd5d93e733919b5c"),
        # Check same
        ({"basis": ""}, "1ae5df953fc38f9b866bd50af6d5513dbfc49cc6"),
        ({"basis": "null"}, "1ae5df953fc38f9b866bd50af6d5513dbfc49cc6"),
        ({"basis": None}, "1ae5df953fc38f9b866bd50af6d5513dbfc49cc6"),
    ],
)  # yapf: disable
def test_result_record_canary_hash(data, hash_index):

    opt = ResultRecord(**{**_base_result, **data})

    assert hash_index == opt.get_hash_index(), data
    assert opt.hash_index is None  # Not set


## Optimization hashes
_qc_spec = {"driver": "gradient", "method": "HF", "basis": "sto-3g", "keywords": None, "program": "prog"}
_base_opt = {"keywords": {}, "program": "prog2", "initial_molecule": "5c7896fb95d592ad07a2fe3b", "qc_spec": _qc_spec}


@pytest.mark.parametrize(
    "data, hash_index",
    [
        # Check same
        ({}, "254de59f1598570d0c31aa2d3d84b601c9da12b9"),
        ({"program": "PROG2"}, "254de59f1598570d0c31aa2d3d84b601c9da12b9"),
        ({"qc_spec": {**_qc_spec, **{"program": "prog"}}}, "254de59f1598570d0c31aa2d3d84b601c9da12b9"),
        ({"qc_spec": {**_qc_spec, **{"method": "HF"}}}, "254de59f1598570d0c31aa2d3d84b601c9da12b9"),
        # Check tolerances
        ({"keywords": {"tol": 1.0e-12}}, "8ab52bff9430f7759323e6a547afc58725422c47"),
        ({"keywords": {"tol": 0.0}}, "8ab52bff9430f7759323e6a547afc58725422c47"),
        ({"keywords": {"tol": 1.0e-9}}, "1628caf9a29c9bf17a66cb55b13106e7f2704e51"),  # Should be different from above
        # Check fields
        ({"initial_molecule": "5c78987e95d592ad07a2fe3c"}, "3c20c8f7b1be857460f2d71d74680dd19e9d9113"),
        # Check basis preps
        ({"qc_spec": {**_qc_spec, **{"basis": None}}}, "3489e0c47144ebedb4fdcc2bfab61f7aa4dc947c"),
        ({"qc_spec": {**_qc_spec, **{"basis": ""}}}, "3489e0c47144ebedb4fdcc2bfab61f7aa4dc947c"),
    ],
)  # yapf: disable
def test_optimization_record_canary_hash(data, hash_index):

    opt = OptimizationRecord(**{**_base_opt, **data})

    assert hash_index == opt.hash_index, data


## GridOptimizationRecord hashes
_opt_spec = {"program": "geometric", "keywords": {"coordsys": "tric"}}

_scan_spec = {"type": "distance", "indices": [1, 2], "steps": [-0.1, 0.0], "step_type": "relative"}  # yapf: disable

_base_gridopt = {
    "keywords": {"preoptimization": False, "scans": [_scan_spec]},
    "optimization_spec": _opt_spec,
    "qc_spec": _qc_spec,
    "initial_molecule": "5c7896fb95d592ad07a2fe3b",
    "starting_molecule": "5c7896fb95d592ad07a2fe3b",
    "grid_optimizations": {},
    "final_energy_dict": {},
    "starting_grid": tuple(),
    "provenance": {"creator": ""},
}


@pytest.mark.parametrize(
    "data, hash_index",
    [
        # Check same
        ({}, "fa2da83aae1651a9115f5eaea83043187c4c8c7b"),
        (
            {
                "keywords": {
                    "preoptimization": False,
                    "scans": [{**_scan_spec, **{"steps": [-0.1 + 1e-12, 0.0 - 1.0e-12]}}],
                }
            },
            "fa2da83aae1651a9115f5eaea83043187c4c8c7b",
        ),
        # Check opt keywords stability
        (
            {"optimization_spec": {**_opt_spec, **{"keywords": {"tol": 1.0e-12}}}},
            "fa2da83aae1651a9115f5eaea83043187c4c8c7b",
        ),
        ({"optimization_spec": {**_opt_spec, **{"keywords": {"tol": 0}}}}, "fa2da83aae1651a9115f5eaea83043187c4c8c7b"),
        # Check fields
        ({"initial_molecule": "5c78987e95d592ad07a2fe3c"}, "9624ce2eca96eabcdb9ec3b2e073429f6dd4b8a4"),
        ({"qc_spec": {**_qc_spec, **{"method": "custom1"}}}, "d6a187bf6de7a25d36402c5d16109ddc6f4f217d"),
    ],
)  # yapf: disable
def test_gridoptimization_canary_hash(data, hash_index):

    gridopt = GridOptimizationRecord(**{**_base_gridopt, **data})

    assert hash_index == gridopt.get_hash_index(), data


## TorsionDriveRecord hashes

_base_torsion = {
    "keywords": {"dihedrals": [[0, 1, 2, 3]], "grid_spacing": [10]},
    "optimization_spec": _opt_spec,
    "qc_spec": _qc_spec,
    "initial_molecule": ["5c7896fb95d592ad07a2fe3b"],
    "final_energy_dict": {},
    "optimization_history": {},
    "minimum_positions": {},
    "provenance": {"creator": ""},
}


@pytest.mark.parametrize(
    "data, hash_index",
    [
        ({}, "dd305011ee2b741b1dcd03350994920a3718b289"),
        # Check same
        (
            {"keywords": {"dihedrals": [[0, 1, 2, 3]], "grid_spacing": [10], "energy_upper_limit": 1.0e-12}},
            "37b65cba19ec4fbd0d54c10fd74d0a27f627cdea",
        ),
        (
            {"keywords": {"dihedrals": [[0, 1, 2, 3]], "grid_spacing": [10], "energy_upper_limit": 0}},
            "37b65cba19ec4fbd0d54c10fd74d0a27f627cdea",
        ),
        (
            {"keywords": {"dihedrals": [[0, 1, 2, 3]], "grid_spacing": [10], "energy_upper_limit": 1.0e-9}},
            "64b400229d3e5bff476e47c093c1a159c69d9fdc",
        ),
        # Check opt keywords stability
        (
            {"optimization_spec": {**_opt_spec, **{"keywords": {"tol": 0.0}}}},
            "a12fd524b0e215b5252b464ca4041091916df8bb",
        ),
        (
            {"optimization_spec": {**_opt_spec, **{"keywords": {"tol": 1.0e-12}}}},
            "a12fd524b0e215b5252b464ca4041091916df8bb",
        ),
        # Check fields
        ({"initial_molecule": ["5c78987e95d592ad07a2fe3c"]}, "f209751a4a6559a8d2d539c070f3b701d1ddf9f2"),
    ],
)  # yapf disable
def test_torsiondrive_canary_hash(data, hash_index):

    td = TorsionDriveRecord(**{**_base_torsion, **data})

    assert hash_index == td.get_hash_index(), data



"""
Queue adapter for Parsl
"""

import logging
import time
import traceback
from typing import Any, Callable, Dict, Hashable, Optional, Tuple

from qcelemental.models import FailedOperation

from .base_adapter import BaseAdapter


def _get_future(future):
    # if future.exception() is None: # This always seems to return None
    try:
        return future.result()
    except Exception as e:
        msg = "Caught Parsl Error:\n" + traceback.format_exc()
        ret = FailedOperation(**{"success": False, "error": {"error_type": e.__class__.__name__, "error_message": msg}})
        return ret


class ParslAdapter(BaseAdapter):
    """An Adapter for Parsl."""

    def __init__(self, client: Any, logger: Optional[logging.Logger] = None, **kwargs):
        BaseAdapter.__init__(self, client, logger, **kwargs)

        import parsl

        self.client = parsl.dataflow.dflow.DataFlowKernel(self.client)
        self._parsl_states = parsl.dataflow.states.States
        self.app_map = {}

    def __repr__(self):
        return "<ParslAdapter client=<DataFlow label='{}'>>".format(self.client.config.executors[0].label)

    def get_app(self, function: str) -> Callable:
        """Obtains a Parsl python_application.

        Parameters
        ----------
        function : str
            A full path to a function

        Returns
        -------
        callable
            The desired AppFactory

        Examples
        --------

        >>> get_app("numpy.einsum")
        <class PythonApp"AppFactory for einsum>
        """

        from parsl.app.app import python_app

        if function in self.app_map:
            return self.app_map[function]

        func = self.get_function(function)

        # TODO set walltime and the like
        self.app_map[function] = python_app(func, data_flow_kernel=self.client)

        return self.app_map[function]

    def _submit_task(self, task_spec: Dict[str, Any]) -> Tuple[Hashable, Any]:

        # Form run tuple
        func = self.get_app(task_spec["spec"]["function"])
        task = func(*task_spec["spec"]["args"], **task_spec["spec"]["kwargs"])
        return task_spec["id"], task

    def count_active_task_slots(self) -> int:

        running = 0
        executor_running_task_map = {key: False for key in self.client.executors.keys()}
        for task in self.queue.values():
            status = self.client.tasks.get(task.tid, {}).get("status", None)
            if status == self._parsl_states.running:
                executor_running_task_map[task["executor"]] = True
            if all(executor_running_task_map.values()):
                # Efficiency loop break
                break

        found_readable = False
        for executor_key, executor in self.client.executors.items():
            if hasattr(executor, "connected_workers"):
                # Should return an int
                running += executor.connected_workers
                found_readable = True
            elif hasattr(executor, "max_threads") and executor_running_task_map[executor_key]:
                running += 1
                found_readable = True

        if not found_readable:
            raise NotImplementedError("Cannot accurately estimate consumption from executors")

        return running

    def acquire_complete(self) -> Dict[str, Any]:
        ret = {}
        del_keys = []
        for key, future in self.queue.items():
            if future.done():
                ret[key] = _get_future(future)
                del_keys.append(key)

        for key in del_keys:
            del self.queue[key]

        return ret

    def await_results(self) -> bool:
        for future in list(self.queue.values()):
            while future.done() is False:
                time.sleep(0.1)

        return True

    def close(self) -> bool:
        self.client.atexit_cleanup()
        return True



"""Msgpack Results Phase 1

Revision ID: 84c94a48e491
Revises: d56ac42b9a43
Create Date: 2019-08-11 17:21:40.264688

"""
from alembic import op
import sqlalchemy as sa
import numpy as np

import os
import sys

sys.path.insert(1, os.path.dirname(os.path.abspath(__file__)))
from migration_helpers import msgpack_migrations
from qcelemental.util import msgpackext_dumps, msgpackext_loads

# revision identifiers, used by Alembic.
revision = "84c94a48e491"
down_revision = "d56ac42b9a43"
branch_labels = None
depends_on = None

block_size = 100
table_name = "result"


def transformer(old_data):

    arr = old_data["return_result"]
    if arr is None:
        pass
    elif old_data["driver"] == "gradient":
        arr = np.array(arr, dtype=float).reshape(-1, 3)
    elif old_data["driver"] == "hessian":
        arr = np.array(arr, dtype=float)
        arr.shape = (-1, int(arr.shape[0] ** 0.5))

    return {"return_result_": msgpackext_dumps(arr)}


update_columns = {"return_result"}


def upgrade():
    msgpack_migrations.json_to_msgpack_table(
        table_name, block_size, update_columns, transformer, read_columns={"driver": sa.String}
    )


def downgrade():
    msgpack_migrations.json_to_msgpack_table_dropcols(table_name, block_size, update_columns)



"""
Base class for computation procedures
"""

import abc
from ..interface.models import KVStore


class BaseTasks(abc.ABC):
    def __init__(self, storage, logger):
        self.storage = storage
        self.logger = logger

    def submit_tasks(self, data):
        """
        Creates results/procedures and tasks in the database
        """

        results_ids, existing_ids = self.parse_input(data)
        submitted_ids = [x for x in results_ids if x not in existing_ids and x is not None]

        n_inserted = 0
        missing = []
        for num, x in enumerate(results_ids):
            if x is None:
                missing.append(num)
            else:
                n_inserted += 1

        results = {
            "meta": {
                "n_inserted": n_inserted,
                "duplicates": [],
                "validation_errors": [],
                "success": True,
                "error_description": False,
                "errors": [],
            },
            "data": {"ids": results_ids, "submitted": submitted_ids, "existing": existing_ids},
        }

        return results

    def retrieve_outputs(self, rdata):
        """
        Retrieves (possibly compressed) outputs from an AtomicResult (that has been converted to a dictionary)

        This function modifies the rdata dictionary in-place
        """

        # Get the compressed outputs if they exist
        stdout = rdata["extras"].pop("_qcfractal_compressed_stdout", None)
        stderr = rdata["extras"].pop("_qcfractal_compressed_stderr", None)
        error = rdata["extras"].pop("_qcfractal_compressed_error", None)

        # Create KVStore objects from these
        if stdout is not None:
            stdout = KVStore(**stdout)
        if stderr is not None:
            stderr = KVStore(**stderr)
        if error is not None:
            error = KVStore(**error)

        # This shouldn't happen, but if they aren't compressed, check for
        # uncompressed
        if stdout is None and rdata.get("stdout", None) is not None:
            self.logger.warning(f"Found uncompressed stdout for result id {rdata['id']}")
            stdout = KVStore(data=rdata["stdout"])
        if stderr is None and rdata.get("stderr", None) is not None:
            self.logger.warning(f"Found uncompressed stderr for result id {rdata['id']}")
            stderr = KVStore(data=rdata["stderr"])
        if error is None and rdata.get("error", None) is not None:
            self.logger.warning(f"Found uncompressed error for result id {rdata['id']}")
            error = KVStore(data=rdata["error"])

        # Now add to the database and set the ids in the diction
        outputs = [stdout, stderr, error]
        stdout_id, stderr_id, error_id = self.storage.add_kvstore(outputs)["data"]
        rdata["stdout"] = stdout_id
        rdata["stderr"] = stderr_id
        rdata["error"] = error_id

    @abc.abstractmethod
    def verify_input(self, data):
        pass

    @abc.abstractmethod
    def parse_input(self, data):
        pass

    @abc.abstractmethod
    def handle_completed_output(self, data):
        pass



"""
Utility functions for QCPortal/QCFractal Interface.
"""
import re
import unicodedata

from pydantic import BaseModel

__all__ = ["replace_dict_keys", "normalize_filename"]


def replace_dict_keys(data, replacement):
    """
    Recursively replaces the keys in data from a dictionary `replacement`.
    """

    if isinstance(data, dict):
        ret = {}
        for k, v in data.items():

            # Replace key
            if k in replacement:
                k = replacement[k]

            # Recurse value if needed.
            new_v = v
            if isinstance(v, dict):
                new_v = replace_dict_keys(v, replacement)
            elif isinstance(v, (list, tuple)):
                new_v = [replace_dict_keys(x, replacement) for x in v]
                if isinstance(v, tuple):
                    new_v = tuple(new_v)

            ret[k] = new_v
        return ret

    elif isinstance(data, (tuple, list)):
        new_data = [replace_dict_keys(x, replacement) for x in data]
        if isinstance(data, tuple):
            new_data = tuple(new_data)

        return new_data

    elif isinstance(data, BaseModel):
        # Handle base model structures
        ret = data.copy()  # Create a copy
        search_keys = data.__fields__.keys()  # Enumerate keys
        for key in search_keys:
            existing_data = getattr(data, key)
            # Try to replace data recursively
            new_data = replace_dict_keys(existing_data, replacement)
            if new_data == existing_data:
                continue  # Do nothing if new data is the same (safer)
            setattr(ret, key, new_data)  # Replace new data in the copy to avoid in-place changes
        return ret  # Return

    else:
        return data


def normalize_filename(value: str) -> str:
    """
    Normalizes string to ASCII, removes non-alpha characters, and converts spaces to underscores.
    """
    value = str(value)
    value = unicodedata.normalize("NFKD", value).encode("ascii", "ignore").decode("ascii")
    value = re.sub(r"[^\w\s-]", "", value).strip()
    return re.sub(r"[-\s]+", "_", value)



import pytest

from . import portal


def test_kvstore_model_fail():
    # data is a string, but compression is not none
    with pytest.raises(ValueError, match=r"Compression is set, but input is a"):
        portal.models.KVStore(**{"data": "123", "compression": "bzip2"})

    # data is a dict, but compression is not none
    with pytest.raises(ValueError, match=r"Compression is set, but input is a"):
        portal.models.KVStore(**{"data": {"123": 123}, "compression": "bzip2"})

    # data is a string, but compression level is not 0
    with pytest.raises(ValueError, match=r"Compression level is set, but input is a"):
        portal.models.KVStore(**{"data": "123", "compression_level": 1})

    # data is a string, but compression level is not 0
    with pytest.raises(ValueError, match=r"Compression level is set, but input is a"):
        portal.models.KVStore(**{"data": {"123": 123}, "compression_level": 1})



"""
Basic ORM models of the QCFractal database

Note: avoid circular import here by including the name of the class
in relations and foreign keys are a string (see TaskQueueORM.base_result_obj)
"""

import datetime

# from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import (
    JSON,
    BigInteger,
    Boolean,
    Column,
    DateTime,
    Enum,
    Float,
    ForeignKey,
    Index,
    Integer,
    LargeBinary,
    String,
)
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship
from sqlalchemy.sql import text

from qcfractal.interface.models.task_models import ManagerStatusEnum, PriorityEnum, TaskStatusEnum
from qcfractal.interface.models.common_models import CompressionEnum
from qcfractal.storage_sockets.models.sql_base import Base, MsgpackExt


class AccessLogORM(Base):
    __tablename__ = "access_log"

    id = Column(Integer, primary_key=True)
    access_date = Column(DateTime, default=datetime.datetime.utcnow, index=True)
    access_method = Column(String, nullable=False)
    access_type = Column(String, nullable=False, index=True)

    # Note: no performance difference between varchar and text in postgres
    # will mostly have a serialized JSON, but not stored as JSON for speed
    extra_params = Column(String)

    # user info
    ip_address = Column(String)
    user_agent = Column(String)

    # extra computed geo data
    city = Column(String)
    country = Column(String)
    country_code = Column(String)
    ip_lat = Column(String)
    ip_long = Column(String)
    postal_code = Column(String)
    subdivision = Column(String)


class ServerStatsLogORM(Base):
    __tablename__ = "server_stats_log"

    id = Column(Integer, primary_key=True)
    timestamp = Column(DateTime, default=datetime.datetime.utcnow, nullable=False)

    # Raw counts
    collection_count = Column(Integer)
    molecule_count = Column(Integer)
    result_count = Column(Integer)
    kvstore_count = Column(Integer)
    access_count = Column(Integer)

    # States
    result_states = Column(JSON)

    # Database
    db_total_size = Column(BigInteger)
    db_table_size = Column(BigInteger)
    db_index_size = Column(BigInteger)
    db_table_information = Column(JSON)

    __table_args__ = (Index("ix_server_stats_log_timestamp", "timestamp"),)


class VersionsORM(Base):
    __tablename__ = "versions"

    id = Column(Integer, primary_key=True)
    created_on = Column(DateTime, default=datetime.datetime.utcnow)
    elemental_version = Column(String, nullable=False)
    fractal_version = Column(String, nullable=False)
    engine_version = Column(String)


class KVStoreORM(Base):
    """TODO: rename to """

    __tablename__ = "kv_store"

    id = Column(Integer, primary_key=True)
    compression = Column(Enum(CompressionEnum), nullable=True)
    compression_level = Column(Integer, nullable=True)
    value = Column(JSON, nullable=True)
    data = Column(LargeBinary, nullable=True)


class MoleculeORM(Base):
    """
    The molecule DB collection is managed by pymongo, so far
    """

    __tablename__ = "molecule"

    id = Column(Integer, primary_key=True)
    molecular_formula = Column(String)

    # TODO - hash can be stored more efficiently (ie, byte array)
    molecule_hash = Column(String)

    # Required data
    schema_name = Column(String)
    schema_version = Column(Integer, default=2)
    symbols = Column(MsgpackExt, nullable=False)
    geometry = Column(MsgpackExt, nullable=False)

    # Molecule data
    name = Column(String, default="")
    identifiers = Column(JSON)
    comment = Column(String)
    molecular_charge = Column(Float, default=0)
    molecular_multiplicity = Column(Integer, default=1)

    # Atom data
    masses = Column(MsgpackExt)
    real = Column(MsgpackExt)
    atom_labels = Column(MsgpackExt)
    atomic_numbers = Column(MsgpackExt)
    mass_numbers = Column(MsgpackExt)

    # Fragment and connection data
    connectivity = Column(JSON)
    fragments = Column(MsgpackExt)
    fragment_charges = Column(JSON)  # Column(ARRAY(Float))
    fragment_multiplicities = Column(JSON)  # Column(ARRAY(Integer))

    # Orientation
    fix_com = Column(Boolean, default=False)
    fix_orientation = Column(Boolean, default=False)
    fix_symmetry = Column(String)

    # Extra
    provenance = Column(JSON)
    extras = Column(JSON)

    __table_args__ = (Index("ix_molecule_hash", "molecule_hash", unique=False),)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class KeywordsORM(Base):
    """
    KeywordsORM are unique for a specific program and name
    """

    __tablename__ = "keywords"

    id = Column(Integer, primary_key=True)
    hash_index = Column(String, nullable=False)
    values = Column(JSON)

    lowercase = Column(Boolean, default=True)
    exact_floats = Column(Boolean, default=False)
    comments = Column(String)

    __table_args__ = (Index("ix_keywords_hash_index", "hash_index", unique=True),)
    # meta = {'indexes': [{'fields': ('hash_index', ), 'unique': True}]}


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class TaskQueueORM(Base):
    """A queue of tasks corresponding to a procedure

    Notes: don't sort query results without having the index sorted
           will impact the performance
    """

    __tablename__ = "task_queue"

    id = Column(Integer, primary_key=True)

    spec = Column(MsgpackExt, nullable=False)

    # others
    tag = Column(String, default=None)
    parser = Column(String, default="")
    program = Column(String)
    procedure = Column(String)
    status = Column(Enum(TaskStatusEnum), default=TaskStatusEnum.waiting)
    priority = Column(Integer, default=int(PriorityEnum.NORMAL))
    manager = Column(String, ForeignKey("queue_manager.name", ondelete="SET NULL"), default=None)

    created_on = Column(DateTime, default=datetime.datetime.utcnow)
    modified_on = Column(DateTime, default=datetime.datetime.utcnow)

    # TODO: for back-compatibility with mongo, tobe removed
    # (requries modifying pydantic model)
    @hybrid_property
    def base_result(self):
        return self.base_result_id

    @base_result.setter
    def base_result(self, val):
        self.base_result_id = int(val)

    # can reference ResultORMs or any ProcedureORM
    base_result_id = Column(Integer, ForeignKey("base_result.id", ondelete="cascade"), unique=True)
    base_result_obj = relationship("BaseResultORM", lazy="select")  # or lazy='joined'

    # An important special case is ORDER BY in combination with LIMIT n: an
    # explicit sort will have to process all the data to identify the first n
    # rows, but if there is an index matching the ORDER BY, the first n rows
    # can be retrieved directly, without scanning the remainder at all.
    __table_args__ = (
        Index("ix_task_queue_created_on", "created_on"),
        Index("ix_task_queue_keys", "status", "program", "procedure", "tag"),
        Index("ix_task_queue_manager", "manager"),
        Index("ix_task_queue_base_result_id", "base_result_id"),
        Index("ix_task_waiting_sort", text("priority desc,  created_on")),
    )


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class ServiceQueueORM(Base):

    __tablename__ = "service_queue"

    id = Column(Integer, primary_key=True)

    status = Column(Enum(TaskStatusEnum), default=TaskStatusEnum.waiting)
    tag = Column(String, default=None)
    hash_index = Column(String, nullable=False)

    procedure_id = Column(Integer, ForeignKey("base_result.id"), unique=True)
    procedure_obj = relationship("BaseResultORM", lazy="joined")

    priority = Column(Integer, default=int(PriorityEnum.NORMAL))
    created_on = Column(DateTime, default=datetime.datetime.utcnow)
    modified_on = Column(DateTime, default=datetime.datetime.utcnow)

    extra = Column(MsgpackExt)

    __table_args__ = (
        Index("ix_service_queue_status", "status"),
        Index("ix_service_queue_priority", "priority"),
        Index("ix_service_queue_modified_on", "modified_on"),
        Index("ix_service_queue_status_tag_hash", "status", "tag"),
        Index("ix_service_queue_hash_index", "hash_index"),
    )


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class UserORM(Base):

    __tablename__ = "user"

    id = Column(Integer, primary_key=True)

    username = Column(String, nullable=False, unique=True)  # indexed and unique
    password = Column(LargeBinary, nullable=False)
    permissions = Column(JSON)  # Column(ARRAY(String))


class QueueManagerLogORM(Base):

    __tablename__ = "queue_manager_logs"

    id = Column(Integer, primary_key=True)
    manager_id = Column(Integer, ForeignKey("queue_manager.id"), nullable=False)

    timestamp = Column(DateTime, default=datetime.datetime.utcnow, nullable=False)

    completed = Column(Integer, nullable=True)
    submitted = Column(Integer, nullable=True)
    failures = Column(Integer, nullable=True)

    total_worker_walltime = Column(Float, nullable=True)
    total_task_walltime = Column(Float, nullable=True)
    active_tasks = Column(Integer, nullable=True)
    active_cores = Column(Integer, nullable=True)
    active_memory = Column(Float, nullable=True)

    __table_args__ = (Index("ix_queue_manager_log_timestamp", "timestamp"),)


class QueueManagerORM(Base):
    """"""

    __tablename__ = "queue_manager"

    id = Column(Integer, primary_key=True)

    name = Column(String, unique=True)
    cluster = Column(String)
    hostname = Column(String)
    username = Column(String)
    uuid = Column(String)
    tag = Column(String)

    # Count at current time
    completed = Column(Integer, default=0)
    submitted = Column(Integer, default=0)
    failures = Column(Integer, default=0)
    returned = Column(Integer, default=0)

    total_worker_walltime = Column(Float, nullable=True)
    total_task_walltime = Column(Float, nullable=True)
    active_tasks = Column(Integer, nullable=True)
    active_cores = Column(Integer, nullable=True)
    active_memory = Column(Float, nullable=True)

    # Adapter Information
    configuration = Column(JSON, nullable=True)

    status = Column(Enum(ManagerStatusEnum), default=ManagerStatusEnum.inactive)

    created_on = Column(DateTime, default=datetime.datetime.utcnow)
    modified_on = Column(DateTime, default=datetime.datetime.utcnow)

    qcengine_version = Column(String)
    manager_version = Column(String)
    programs = Column(JSON)
    procedures = Column(JSON)

    __table_args__ = (Index("ix_queue_manager_status", "status"), Index("ix_queue_manager_modified_on", "modified_on"))



"""Final migration synchronization from v0.11 to v0.12

Revision ID: 4b27843a188a
Revises: 159ba85908fd
Create Date: 2019-11-06 13:48:39.716633

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "4b27843a188a"
down_revision = "159ba85908fd"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "collection", "group", existing_type=sa.VARCHAR(), type_=sa.String(length=100), existing_nullable=False
    )
    op.alter_column("molecule", "geometry", existing_type=postgresql.BYTEA(), nullable=False)
    op.alter_column("molecule", "symbols", existing_type=postgresql.BYTEA(), nullable=False)
    op.alter_column("task_queue", "spec", existing_type=postgresql.BYTEA(), nullable=False)
    op.drop_constraint("task_queue_manager_fkey", "task_queue", type_="foreignkey")
    op.create_foreign_key(
        "task_queue_manager_fkey", "task_queue", "queue_manager", ["manager"], ["name"], ondelete="SET NULL"
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint("task_queue_manager_fkey", "task_queue", type_="foreignkey")
    op.create_foreign_key("task_queue_manager_fkey", "task_queue", "queue_manager", ["manager"], ["name"])
    op.alter_column("task_queue", "spec", existing_type=postgresql.BYTEA(), nullable=True)
    op.alter_column("molecule", "symbols", existing_type=postgresql.BYTEA(), nullable=True)
    op.alter_column("molecule", "geometry", existing_type=postgresql.BYTEA(), nullable=True)
    op.alter_column(
        "collection", "group", existing_type=sa.String(length=100), type_=sa.VARCHAR(), existing_nullable=False
    )
    # ### end Alembic commands ###



"""Adds a empty protocols column to `base_result` and wavefunction to `result`

Revision ID: c194a8ef6acf
Revises: cc54842bb6ba
Create Date: 2019-09-30 10:31:23.552722

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "c194a8ef6acf"
down_revision = "cc54842bb6ba"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("base_result", sa.Column("protocols", postgresql.JSONB(astext_type=sa.Text()), nullable=True))
    op.execute("UPDATE base_result SET protocols='{}'::json")
    op.alter_column("base_result", "protocols", nullable=False)
    op.add_column("result", sa.Column("wavefunction", postgresql.JSONB(astext_type=sa.Text()), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("base_result", "protocols")
    op.drop_column("result", "wavefunction")
    # ### end Alembic commands ###



"""Msgpack Molecule Phase 2

Revision ID: d56ac42b9a43
Revises: 963822c28879
Create Date: 2019-08-11 16:17:23.856255

"""
from alembic import op
import sqlalchemy as sa

import os
import sys

sys.path.insert(1, os.path.dirname(os.path.abspath(__file__)))

from migration_helpers import msgpack_migrations

# revision identifiers, used by Alembic.
revision = "d56ac42b9a43"
down_revision = "963822c28879"
branch_labels = None
depends_on = None

table_name = "molecule"
update_columns = {"symbols", "geometry", "masses", "real", "atom_labels", "atomic_numbers", "mass_numbers", "fragments"}

nullable = update_columns.copy()
nullable -= {"symbols", "geometry"}


def upgrade():
    msgpack_migrations.json_to_msgpack_table_altercolumns(table_name, update_columns, nullable_true=nullable)


def downgrade():
    raise ValueError("Cannot downgrade json to msgpack conversions")



"""
QCPortal Database ODM
"""
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple, Union

import pandas as pd

from ..models import ObjectId, OptimizationSpecification, ProtoModel, QCSpecification, TorsionDriveInput
from ..models.torsiondrive import TDKeywords
from ..visualization import custom_plot
from .collection import BaseProcedureDataset
from .collection_utils import register_collection

if TYPE_CHECKING:  # pragma: no cover
    from ..models import Molecule


class TDEntry(ProtoModel):
    """Data model for the `reactions` list in Dataset"""

    name: str
    initial_molecules: Set[ObjectId]
    td_keywords: TDKeywords
    attributes: Dict[str, Any]
    object_map: Dict[str, ObjectId] = {}


class TDEntrySpecification(ProtoModel):
    name: str
    description: Optional[str]
    optimization_spec: OptimizationSpecification
    qc_spec: QCSpecification


class TorsionDriveDataset(BaseProcedureDataset):
    class DataModel(BaseProcedureDataset.DataModel):

        records: Dict[str, TDEntry] = {}
        history: Set[str] = set()
        specs: Dict[str, TDEntrySpecification] = {}

        class Config(BaseProcedureDataset.DataModel.Config):
            pass

    def _internal_compute_add(self, spec: Any, entry: Any, tag: str, priority: str) -> ObjectId:

        service = TorsionDriveInput(
            initial_molecule=entry.initial_molecules,
            keywords=entry.td_keywords,
            optimization_spec=spec.optimization_spec,
            qc_spec=spec.qc_spec,
        )

        return self.client.add_service([service], tag=tag, priority=priority).ids[0]

    def add_specification(
        self,
        name: str,
        optimization_spec: OptimizationSpecification,
        qc_spec: QCSpecification,
        description: Optional[str] = None,
        overwrite: bool = False,
    ) -> None:
        """
        Parameters
        ----------
        name : str
            The name of the specification
        optimization_spec : OptimizationSpecification
            A full optimization specification for TorsionDrive
        qc_spec : QCSpecification
            A full quantum chemistry specification for TorsionDrive
        description : str, optional
            A short text description of the specification
        overwrite : bool, optional
            Overwrite existing specification names

        """

        spec = TDEntrySpecification(
            name=name, optimization_spec=optimization_spec, qc_spec=qc_spec, description=description
        )

        return self._add_specification(name, spec, overwrite=overwrite)

    def add_entry(
        self,
        name: str,
        initial_molecules: List["Molecule"],
        dihedrals: List[Tuple[int, int, int, int]],
        grid_spacing: List[int],
        dihedral_ranges: Optional[List[Tuple[int, int]]] = None,
        energy_decrease_thresh: Optional[float] = None,
        energy_upper_limit: Optional[float] = None,
        attributes: Dict[str, Any] = None,
        save: bool = True,
    ) -> None:
        """
        Parameters
        ----------
        name : str
            The name of the entry, will be used for the index
        initial_molecules : List[Molecule]
            The list of starting Molecules for the TorsionDrive
        dihedrals : List[Tuple[int, int, int, int]]
            A list of dihedrals to scan over
        grid_spacing : List[int]
            The grid spacing for each dihedrals
        dihedral_ranges: Optional[List[Tuple[int, int]]]
            The range limit of each dihedrals to scan, within [-180, 360]
        energy_decrease_thresh: Optional[float]
            The threshold of energy decrease to trigger activating grid points
        energy_upper_limit: Optional[float]
            The upper limit of energy relative to current global minimum to trigger activating grid points
        attributes : Dict[str, Any], optional
            Additional attributes and descriptions for the entry
        save : bool, optional
            If true, saves the collection after adding the entry. If this is False be careful
            to call save after all entries are added, otherwise data pointers may be lost.
        """

        self._check_entry_exists(name)  # Fast skip

        if attributes is None:
            attributes = {}

        # Build new objects
        molecule_ids = self.client.add_molecules(initial_molecules)
        td_keywords = TDKeywords(
            dihedrals=dihedrals,
            grid_spacing=grid_spacing,
            dihedral_ranges=dihedral_ranges,
            energy_decrease_thresh=energy_decrease_thresh,
            energy_upper_limit=energy_upper_limit,
        )

        entry = TDEntry(name=name, initial_molecules=molecule_ids, td_keywords=td_keywords, attributes=attributes)

        self._add_entry(name, entry, save)

    def counts(
        self,
        entries: Union[str, List[str]],
        specs: Optional[Union[str, List[str]]] = None,
        count_gradients: bool = False,
    ) -> pd.DataFrame:
        """Counts the number of optimization or gradient evaluations associated with the
        TorsionDrives.

        Parameters
        ----------
        entries : Union[str, List[str]]
            The entries to query for
        specs : Optional[Union[str, List[str]]], optional
            The specifications to query for
        count_gradients : bool, optional
            If True, counts the total number of gradient calls. Warning! This can be slow for large datasets.

        Returns
        -------
        DataFrame
            The queried counts.
        """

        if isinstance(specs, str):
            specs = [specs]

        if isinstance(entries, str):
            entries = [entries]

        # Query all of the specs and make sure they are valid
        if specs is None:
            specs = list(self.df.columns)
        else:
            new_specs = []
            for spec in specs:
                new_specs.append(self.query(spec))

            # Remap names
            specs = new_specs

        # Count functions
        def count_gradient_evals(td):
            if td.status != "COMPLETE":
                return None

            total_grads = 0
            for key, optimizations in td.get_history().items():
                for opt in optimizations:
                    total_grads += len(opt.trajectory)
            return total_grads

        def count_optimizations(td):
            if td.status != "COMPLETE":
                return None
            return sum(len(v) for v in td.optimization_history.values())

        # Loop over the data and apply the count function
        ret = []
        for col in specs:
            data = self.df[col]
            if entries:
                data = data[entries]

            if count_gradients:
                cnts = data.apply(lambda td: count_gradient_evals(td))
            else:
                cnts = data.apply(lambda td: count_optimizations(td))
            ret.append(cnts)

        ret = pd.DataFrame(ret).transpose()
        ret.dropna(inplace=True, how="all")
        # ret = pd.DataFrame([ret[x].astype(int) for x in ret.columns]).transpose()
        return ret

    def visualize(
        self,
        entries: Union[str, List[str]],
        specs: Union[str, List[str]],
        relative: bool = True,
        units: str = "kcal / mol",
        digits: int = 3,
        use_measured_angle: bool = False,
        return_figure: Optional[bool] = None,
    ) -> "plotly.Figure":
        """
        Parameters
        ----------
        entries : Union[str, List[str]]
            A single or list of indices to plot.
        specs : Union[str, List[str]]
            A single or list of specifications to plot.
        relative : bool, optional
            Shows relative energy, lowest energy per scan is zero.
        units : str, optional
            The units of the plot.
        digits : int, optional
            Rounds the energies to n decimal places for display.
        use_measured_angle : bool, optional
            If True, the measured final angle instead of the constrained optimization angle.
            Can provide more accurate results if the optimization was ill-behaved,
            but pulls additional data from the server and may take longer.
        return_figure : Optional[bool], optional
            If True, return the raw plotly figure. If False, returns a hosted iPlot. If None, return a iPlot display in Jupyter notebook and a raw plotly figure in all other circumstances.

        Returns
        -------
        plotly.Figure
            The requested figure.
        """

        show_spec = True
        if isinstance(specs, str):
            specs = [specs]
            show_spec = False

        if isinstance(entries, str):
            entries = [entries]

        # Query all of the specs and make sure they are valid
        formatted_spec_names = []
        for spec in specs:
            formatted_spec_names.append(self.query(spec))

        traces = []
        ranges = []
        # Loop over specifications
        for spec in formatted_spec_names:
            # Loop over indices (groups colors by entry)
            for index in entries:

                # Plot the figure using the torsiondrives plotting function
                fig = self.df.loc[index, spec].visualize(
                    relative=relative,
                    units=units,
                    digits=digits,
                    use_measured_angle=use_measured_angle,
                    return_figure=True,
                )

                ranges.append(fig.layout.xaxis.range)
                trace = fig.data[0]  # Pull out the underlying scatterplot

                if show_spec:
                    trace.name = f"{index}-{spec}"
                else:
                    trace.name = f"{index}"

                traces.append(trace)

        title = "TorsionDriveDataset 1-D Plot"
        if show_spec is False:
            title += f" [spec={formatted_spec_names[0]}]"

        if relative:
            ylabel = f"Relative Energy [{units}]"
        else:
            ylabel = f"Absolute Energy [{units}]"

        custom_layout = {
            "title": title,
            "yaxis": {"title": ylabel, "zeroline": True},
            "xaxis": {
                "title": "Dihedral Angle [degrees]",
                "zeroline": False,
                "range": [min(x[0] for x in ranges), max(x[1] for x in ranges)],
            },
        }

        return custom_plot(traces, custom_layout, return_figure=return_figure)


register_collection(TorsionDriveDataset)



"""
Wraps geometric procedures
"""

import io
import copy
import json
import contextlib
from typing import Any, Dict, List

import numpy as np

from ..extras import find_module
from ..interface.models import TorsionDriveRecord
from .service_util import BaseService, TaskManager

__all__ = ["TorsionDriveService"]

__td_api = find_module("torsiondrive")


def _check_td():
    if __td_api is None:
        raise ModuleNotFoundError(
            "Unable to find TorsionDriveRecord which must be installed to use the TorsionDriveService"
        )


class TorsionDriveService(BaseService):

    # Index info
    service: str = "torsiondrive"
    program: str = "torsiondrive"
    procedure: str = "torsiondrive"

    # Program info
    optimization_program: str

    # Output
    output: TorsionDriveRecord = None  # added default

    # Temporaries
    torsiondrive_state: Dict[str, Any]
    optimization_history: Dict[str, List[str]] = {}

    # Task helpers
    task_map: Dict[str, List[str]] = {}
    task_manager: TaskManager = TaskManager()

    # Templates
    dihedral_template: str
    optimization_template: str
    molecule_template: str

    @classmethod
    def initialize_from_api(cls, storage_socket, logger, service_input, tag=None, priority=None):
        _check_td()
        import torsiondrive
        from torsiondrive import td_api

        # Build the record
        # TODO: This removes the task_id which may exist on old records, but does not exist
        # in newer TorsionDriveRecords.
        # If a proper migration is ever done,
        output = TorsionDriveRecord(
            **service_input.dict(exclude={"initial_molecule", "task_id"}),
            initial_molecule=[x.id for x in service_input.initial_molecule],
            provenance={
                "creator": "torsiondrive",
                "version": torsiondrive.__version__,
                "routine": "torsiondrive.td_api",
            },
            final_energy_dict={},
            minimum_positions={},
            optimization_history={},
        )

        meta = {"output": output}

        # Remove identity info from molecule template
        molecule_template = copy.deepcopy(service_input.initial_molecule[0].dict(encoding="json"))
        molecule_template.pop("id", None)
        molecule_template.pop("identifiers", None)
        meta["molecule_template"] = json.dumps(molecule_template)

        # Initiate torsiondrive meta
        # The torsiondrive package uses print, so capture that using
        # contextlib
        td_stdout = io.StringIO()
        with contextlib.redirect_stdout(td_stdout):
            meta["torsiondrive_state"] = td_api.create_initial_state(
                dihedrals=output.keywords.dihedrals,
                grid_spacing=output.keywords.grid_spacing,
                elements=molecule_template["symbols"],
                init_coords=[x.geometry for x in service_input.initial_molecule],
                dihedral_ranges=output.keywords.dihedral_ranges,
                energy_decrease_thresh=output.keywords.energy_decrease_thresh,
                energy_upper_limit=output.keywords.energy_upper_limit,
            )

        meta["stdout"] = td_stdout.getvalue()

        # Build dihedral template
        dihedral_template = []
        for idx in output.keywords.dihedrals:
            tmp = {"type": "dihedral", "indices": idx}
            dihedral_template.append(tmp)

        meta["dihedral_template"] = json.dumps(dihedral_template)

        # Build optimization template
        opt_template = {
            "meta": {"procedure": "optimization", "qc_spec": output.qc_spec.dict(), "tag": meta.pop("tag", None)}
        }
        opt_template["meta"].update(output.optimization_spec.dict())
        meta["optimization_template"] = json.dumps(opt_template)

        # Move around geometric data
        meta["optimization_program"] = output.optimization_spec.program

        meta["hash_index"] = output.get_hash_index()

        meta["task_tag"] = tag
        meta["task_priority"] = priority
        return cls(**meta, storage_socket=storage_socket, logger=logger)

    def iterate(self):
        _check_td()
        from torsiondrive import td_api

        self.status = "RUNNING"

        # Check if tasks are done
        if self.task_manager.done() is False:
            return False

        complete_tasks = self.task_manager.get_tasks()

        # Populate task results
        task_results = {}
        for key, task_ids in self.task_map.items():
            task_results[key] = []

            for task_id in task_ids:
                # Cycle through all tasks for this entry
                ret = complete_tasks[task_id]

                # Lookup molecules
                initial_id = ret["initial_molecule"]
                final_id = ret["final_molecule"]

                mol_ids = [initial_id, final_id]
                mol_data = self.storage_socket.get_molecules(id=mol_ids)["data"]
                mol_map = {x.id: x.geometry for x in mol_data}

                task_results[key].append((mol_map[initial_id], mol_map[final_id], ret["energies"][-1]))

        # The torsiondrive package uses print, so capture that using
        # contextlib
        td_stdout = io.StringIO()
        with contextlib.redirect_stdout(td_stdout):
            td_api.update_state(self.torsiondrive_state, task_results)

            # Create new tasks from the current state
            next_tasks = td_api.next_jobs_from_state(self.torsiondrive_state, verbose=True)

        self.stdout += "\n" + td_stdout.getvalue()

        # All done
        if len(next_tasks) == 0:
            self.status = "COMPLETE"
            self.update_output()
            return True

        self.submit_optimization_tasks(next_tasks)

        return False

    def submit_optimization_tasks(self, task_dict):
        _check_td()
        from torsiondrive import td_api

        new_tasks = {}
        task_map = {}

        for key, geoms in task_dict.items():
            task_map[key] = []
            for num, geom in enumerate(geoms):

                # Update molecule
                packet = json.loads(self.optimization_template)

                # Construct constraints
                constraints = json.loads(self.dihedral_template)
                grid_id = td_api.grid_id_from_string(key)
                for con_num, k in enumerate(grid_id):
                    constraints[con_num]["value"] = k
                # update existing constraints to support the "extra constraints" feature
                packet["meta"]["keywords"].setdefault("constraints", {})
                packet["meta"]["keywords"]["constraints"].setdefault("set", [])
                packet["meta"]["keywords"]["constraints"]["set"].extend(constraints)

                # Build new molecule
                mol = json.loads(self.molecule_template)
                mol["geometry"] = geom
                packet["data"] = [mol]

                task_key = "{}-{}".format(key, num)
                new_tasks[task_key] = packet

                task_map[key].append(task_key)

        self.task_manager.submit_tasks("optimization", new_tasks)
        self.task_map = task_map

        # Update history
        for key, task_ids in self.task_map.items():
            if key not in self.optimization_history:
                self.optimization_history[key] = []

            for task_id in task_ids:
                self.optimization_history[key].append(self.task_manager.required_tasks[task_id])

        self.update_output()

    def update_output(self):
        """
        Adds data to the TorsionDriveRecord object
        """
        _check_td()
        from torsiondrive import td_api

        # # Get lowest energies and positions
        min_positions = {}
        final_energy = {}
        for k, v in self.torsiondrive_state["grid_status"].items():
            idx = int(np.argmin([x[2] for x in v]))
            key = json.dumps(td_api.grid_id_from_string(k))
            min_positions[key] = idx
            final_energy[key] = v[idx][2]

        history = {json.dumps(td_api.grid_id_from_string(k)): v for k, v in self.optimization_history.items()}

        self.output = self.output.copy(
            update={
                "status": self.status,
                "minimum_positions": min_positions,
                "final_energy_dict": final_energy,
                "optimization_history": history,
            }
        )

        return True



"""
A model for GridOptimization
"""
import copy
import json
from enum import Enum
from typing import Any, Dict, List, Tuple, Union

from pydantic import Field, constr, validator

from .common_models import Molecule, ObjectId, OptimizationSpecification, ProtoModel, QCSpecification
from .model_utils import recursive_normalizer
from .records import RecordBase

__all__ = ["GOKeywords", "GridOptimizationInput", "GridOptimizationRecord", "ScanDimension"]


class ScanTypeEnum(str, Enum):
    """
    The type of scan to perform. This choices is limited to the scan types allowed by the scan dimensions.
    """

    distance = "distance"
    angle = "angle"
    dihedral = "dihedral"


class StepTypeEnum(str, Enum):
    """
    The types of steps to take in a scan dimension: either in absolute or relative terms. ``relative`` indicates that
    the values are relative to the starting value (e.g., a bond starts as 2.1 Bohr, relative steps of [-0.1, 0, 1.0]
    indicate grid points of [2.0, 2.1, 3.1] Bohr. An ``absolute`` ``step_type`` will be exactly those values instead."
    """

    absolute = "absolute"
    relative = "relative"


class ScanDimension(ProtoModel):
    """
    A full description of a dimension to scan over.
    """

    type: ScanTypeEnum = Field(..., description=str(ScanTypeEnum.__doc__))
    indices: List[int] = Field(
        ...,
        description="The indices of atoms to select for the scan. The size of this is a function of the type. e.g., "
        "distances, angles and dihedrals require 2, 3, and 4 atoms, respectively.",
    )
    steps: List[float] = Field(
        ...,
        description="Step sizes to scan in relative to your current location in the scan. This must be a strictly "
        "monotonic series.",
        units=["Bohr", "degrees"],
    )
    step_type: StepTypeEnum = Field(..., description=str(StepTypeEnum.__doc__))

    @validator("type", "step_type", pre=True)
    def check_lower_type_step_type(cls, v):
        return v.lower()

    @validator("indices")
    def check_indices(cls, v, values, **kwargs):
        sizes = {ScanTypeEnum.distance: 2, ScanTypeEnum.angle: 3, ScanTypeEnum.dihedral: 4}
        if sizes[values["type"]] != len(v):
            raise ValueError(
                "ScanDimension of type {} must have {} values, found {}.".format(
                    values["type"], sizes[values["type"]], len(v)
                )
            )

        return v

    @validator("steps")
    def check_steps(cls, v):
        if not (all(x < y for x, y in zip(v, v[1:])) or all(x > y for x, y in zip(v, v[1:]))):
            raise ValueError("Steps are not strictly monotonically increasing or decreasing.")

        v = recursive_normalizer(v)

        return v


class GOKeywords(ProtoModel):
    """
    GridOptimizationRecord options.
    """

    scans: List[ScanDimension] = Field(
        ..., description="The dimensions to scan along (along with their options) for the GridOptimization."
    )
    preoptimization: bool = Field(
        True,
        description="If ``True``, first runs an unrestricted optimization before starting the grid computations. "
        "This is especially useful when combined with ``relative`` ``step_types``.",
    )


_gridopt_constr = constr(strip_whitespace=True, regex="gridoptimization")
_qcfractal_constr = constr(strip_whitespace=True, regex="qcfractal")


class GridOptimizationInput(ProtoModel):
    """
    The input to create a GridOptimization Service with.

    """

    program: _qcfractal_constr = Field(
        "qcfractal",
        description="The name of the source program which initializes the Grid Optimization. This is a constant "
        "and is used for provenance information.",
    )
    procedure: _gridopt_constr = Field(
        "gridoptimization",
        description="The name of the procedure being run. This is a constant and is used for provenance information.",
    )
    initial_molecule: Union[ObjectId, Molecule] = Field(
        ...,
        description="The Molecule to begin the Grid Optimization with. This can either be an existing Molecule in "
        "the database (through its :class:`ObjectId`) or a fully specified :class:`Molecule` model.",
    )
    keywords: GOKeywords = Field(..., description="The keyword options to run the Grid Optimization.")
    optimization_spec: OptimizationSpecification = Field(
        ..., description="The specification to run the underlying optimization through at each grid point."
    )
    qc_spec: QCSpecification = Field(
        ...,
        description="The specification for each of the quantum chemistry calculations run in each geometry "
        "optimization.",
    )


class GridOptimizationRecord(RecordBase):
    """
    The record of a GridOptimization service result.

    A GridOptimization is a type of constrained optimization in which a set of dimension are scanned over. An
    is to compute the

    """

    # Classdata
    _hash_indices = {"initial_molecule", "keywords", "optimization_meta", "qc_spec"}

    # Version data
    version: int = Field(1, description="The version number of the Record.")
    procedure: _gridopt_constr = Field(
        "gridoptimization",
        description="The name of the procedure being run, which is Grid Optimization. This is a constant "
        "and is used for provenance information.",
    )
    program: _qcfractal_constr = Field(
        "qcfractal",
        description="The name of the source program which initializes the Grid Optimization. This is a constant "
        "and is used for provenance information.",
    )

    # Input data
    initial_molecule: ObjectId = Field(..., description="Id of the initial molecule in the database.")
    keywords: GOKeywords = Field(..., description="The keywords for this Grid Optimization.")
    optimization_spec: OptimizationSpecification = Field(
        ..., description="The specification of each geometry optimization."
    )
    qc_spec: QCSpecification = Field(
        ...,
        description="The specification for each of the quantum chemistry computations used by the geometry "
        "optimizations.",
    )

    # Output data
    starting_molecule: ObjectId = Field(
        ...,
        description="Id of the molecule in the database begins the grid optimization. "
        "This will differ from the ``initial_molecule`` if ``preoptimization`` is True.",
    )
    final_energy_dict: Dict[str, float] = Field(
        ..., description="Map of the final energy from the grid optimization at each grid point."
    )
    grid_optimizations: Dict[str, ObjectId] = Field(..., description="The Id of each optimization at each grid point.")
    starting_grid: tuple = Field(
        ...,
        description="Initial grid point from which the Grid Optimization started. This grid point is the closest in "
        "structure to the ``starting_molecule``.",
    )  # yapf: disable

    ## Utility

    def _organize_return(self, data: Dict[str, Any], key: Union[int, str, None]) -> Dict[str, Any]:

        if key is None:
            return {self.deserialize_key(k): copy.deepcopy(v) for k, v in data.items()}
        else:
            key = self.serialize_key(key)

        return copy.deepcopy(data[key])

    @staticmethod
    def serialize_key(key: Union[int, Tuple[int]]) -> str:
        """Serializes the key to map to the internal keys.

        Parameters
        ----------
        key : Union[int, Tuple[int]]
            A integer or list of integers denoting the position in the grid
            to find.

        Returns
        -------
        str
            The internal key value.
        """
        if isinstance(key, (int, float)):
            key = (int(key),)

        return json.dumps(key)

    @staticmethod
    def deserialize_key(key: str) -> Tuple[int]:
        """Unpacks a string key to a python object.

        Parameters
        ----------
        key : str
            The input key

        Returns
        -------
        Tuple[int]
            The unpacked key.
        """
        data = json.loads(key)
        if data == "preoptimization":
            return data
        else:
            return tuple(data)

    def get_scan_value(self, scan_number: Union[str, int, Tuple[int]]) -> Tuple[float, ...]:
        """
        Obtains the scan parameters at a given grid point.

        Parameters
        ----------
        scan_number : Union[str, int, Tuple[int]]
            The key of the scan.

        Returns
        -------
        Tuple[float, ...]
            Description
        """
        if isinstance(scan_number, str):
            scan_number = self.deserialize_key(scan_number)

        ret = []
        for n, idx in enumerate(scan_number):
            ret.append(self.keywords.scans[n].steps[idx])

        return tuple(ret)

    def get_scan_dimensions(self) -> Tuple[float, ...]:
        """
        Returns the overall dimensions of the scan.

        Returns
        -------
        Tuple[float, ...]
            The size of each dimension in the scan.
        """
        ret = []
        for scan in self.keywords.scans:
            ret.append(len(scan.steps))

        return tuple(ret)

    def detailed_status(self) -> Dict[str, Any]:

        # Compute the total number of grid points
        tpoints = 1
        for scan in self.keywords.scans:
            tpoints *= len(scan.steps)

        if self.keywords.preoptimization:
            tpoints += 1

        flat_history = list(self.get_history().values())

        ret = {
            "status": self.status.value,
            "total_points": tpoints,
            "computed_points": len(self.grid_optimizations),
            "complete_tasks": sum(x.status == "COMPLETE" for x in flat_history),
            "incomplete_tasks": sum((x.status == "INCOMPLETE") or (x.status == "RUNNING") for x in flat_history),
            "error_tasks": sum(x.status == "ERROR" for x in flat_history),
        }
        ret["current_tasks"] = ret["error_tasks"] + ret["incomplete_tasks"]
        ret["percent_complete"] = ret["computed_points"] / ret["total_points"] * 100
        ret["errors"] = [x for x in flat_history if x.status == "ERROR"]

        return ret

    ## Query

    def get_history(self, key: Union[int, str, None] = None) -> Dict[str, "Optimization"]:
        """Pulls the optimization history of the computation.

        Parameters
        ----------
        key : Union[int, str, None], optional
            Specifies a single entry to pull from.

        Returns
        -------
        Dict[str, 'Optimization']
            Return the optimizations in the computed history.
        """

        if "optimization_history" not in self.cache:
            procs = self.client.query_procedures(id=list(self.grid_optimizations.values()))
            proc_map = {x.id: x for x in procs}

            self.cache["optimization_history"] = {k: proc_map[v] for k, v in self.grid_optimizations.items()}

        return self._organize_return(self.cache["optimization_history"], key)

    def get_final_energies(self, key: Union[int, str, None] = None) -> Dict[str, float]:
        """
        Provides the final optimized energies at each grid point.

        Parameters
        ----------
        key : Union[int, str, None], optional
            Specifies a single entry to pull from.

        Returns
        -------
        energy : Dict[str, float]
            Returns energies at each grid point in a dictionary or at a
            single point if a key is specified.

        Examples
        --------
        >>> grid_optimization_record.get_final_energies()
        {(-90,): -148.7641654446243, (180,): -148.76501336993732, (0,): -148.75056290106735, (90,): -148.7641654446148}

        >>> grid_optimization_record.get_final_energies((-90,))
        -148.7641654446243

        """

        return self._organize_return(self.final_energy_dict, key)

    def get_final_molecules(self, key: Union[int, str, None] = None) -> Dict[str, "Molecule"]:
        """
        Provides the final optimized molecules at each grid point.

        Parameters
        ----------
        key : Union[int, str, None], optional
            Specifies a single entry to pull from.


        Returns
        -------
        final_molecules : Dict[str, 'Molecule']
            Returns energies at each grid point in a dictionary or at a
            single point if a key is specified.

        Examples
        --------

        >>> mols = grid_optimization_record.get_final_molecules()
        >>> type(mols[(-90, )])
        qcelemental.models.molecule.Molecule

        >>> type(grid_optimization_record.get_final_molecules((-90,)))
        qcelemental.models.molecule.Molecule

        """

        if "final_molecules" not in self.cache:
            ret = {}
            for k, task_id in self.grid_optimizations.items():
                task = self.client.query_procedures(id=task_id)[0]
                ret[k] = task.get_final_molecule()
            self.cache["final_molecules"] = ret

        data = self.cache["final_molecules"]
        return self._organize_return(data, key)

    def get_final_results(self, key: Union[int, Tuple[int, ...], str] = None) -> Dict[str, "ResultRecord"]:
        """Returns the final opt gradient result records at each grid point.

        Parameters
        ----------
        key : Union[int, Tuple[int, ...], str], optional
            Specifies a single entry to pull from.

        Returns
        -------
        final_results : Dict[str, 'ResultRecord']
            Returns ResultRecord at each grid point in a dictionary or at a
            single point if a key is specified.

        Examples
        --------
        >>> mols = grid_optimization_record.get_final_results()
        >>> type(mols[(-90, )])
        qcfractal.interface.models.records.ResultRecord

        >>> type(grid_optimization_record.get_final_results((-90,)))
        qcfractal.interface.models.records.ResultRecord

        """

        if "final_results" not in self.cache:
            map_id_key = {}
            ret = {}
            for k, task_id in self.grid_optimizations.items():
                task = self.client.query_procedures(id=task_id)[0]
                if len(task.trajectory) > 0:
                    final_grad_record_id = task.trajectory[-1]
                    # store the id -> grid id mapping
                    map_id_key[final_grad_record_id] = k
            # combine the ids into one query
            query_result_ids = list(map_id_key.keys())
            # run the query on this batch
            for grad_result_record in self.client.query_results(id=query_result_ids):
                k = map_id_key[grad_result_record.id]
                ret[k] = grad_result_record

            self.cache["final_results"] = ret

        data = self.cache["final_results"]

        return self._organize_return(data, key)



"""
The FractalServer class
"""

import asyncio
import datetime
import logging
import ssl
import time
import traceback
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Dict, List, Optional, Union
from qcelemental.models import ComputeError

import tornado.ioloop
import tornado.log
import tornado.options
import tornado.web

from .extras import get_information
from .interface import FractalClient
from .queue import QueueManager, QueueManagerHandler, ServiceQueueHandler, TaskQueueHandler, ComputeManagerHandler
from .services import construct_service
from .storage_sockets import ViewHandler, storage_socket_factory
from .storage_sockets.api_logger import API_AccessLogger
from .web_handlers import (
    CollectionHandler,
    InformationHandler,
    KeywordHandler,
    KVStoreHandler,
    MoleculeHandler,
    OptimizationHandler,
    ProcedureHandler,
    ResultHandler,
    WavefunctionStoreHandler,
)


def _build_ssl():
    from cryptography import x509
    from cryptography.x509.oid import NameOID
    from cryptography.hazmat.primitives import hashes
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.primitives.asymmetric import rsa

    import sys
    import socket
    import ipaddress
    import random

    hostname = socket.gethostname()
    public_ip = ipaddress.ip_address(socket.gethostbyname(hostname))

    key = rsa.generate_private_key(public_exponent=65537, key_size=1024, backend=default_backend())

    alt_name_list = [x509.DNSName(hostname), x509.IPAddress(ipaddress.ip_address(public_ip))]
    alt_names = x509.SubjectAlternativeName(alt_name_list)

    # Basic data
    name = x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, hostname)])
    basic_contraints = x509.BasicConstraints(ca=True, path_length=0)
    now = datetime.datetime.utcnow()

    # Build cert
    cert = (
        x509.CertificateBuilder()
        .subject_name(name)
        .issuer_name(name)
        .public_key(key.public_key())
        .serial_number(int(random.random() * sys.maxsize))
        .not_valid_before(now)
        .not_valid_after(now + datetime.timedelta(days=10 * 365))
        .add_extension(basic_contraints, False)
        .add_extension(alt_names, False)
        .sign(key, hashes.SHA256(), default_backend())
    )  # yapf: disable

    # Build and return keys
    cert_pem = cert.public_bytes(encoding=serialization.Encoding.PEM)
    key_pem = key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    )  # yapf: disable

    return cert_pem, key_pem


class FractalServer:
    def __init__(
        self,
        # Server info options
        name: str = "QCFractal Server",
        port: int = 7777,
        loop: "IOLoop" = None,
        compress_response: bool = True,
        # Security
        security: Optional[str] = None,
        allow_read: bool = False,
        ssl_options: Union[bool, Dict[str, str]] = True,
        # Database options
        storage_uri: str = "postgresql://localhost:5432",
        storage_project_name: str = "qcfractal_default",
        query_limit: int = 1000,
        # View options
        view_enabled: bool = False,
        view_path: Optional[str] = None,
        # Log options
        logfile_prefix: str = None,
        loglevel: str = "info",
        log_apis: bool = False,
        geo_file_path: str = None,
        # Queue options
        queue_socket: "BaseAdapter" = None,
        heartbeat_frequency: float = 1800,
        # Service options
        max_active_services: int = 20,
        service_frequency: float = 60,
        # Testing functions
        skip_storage_version_check=True,
    ):
        """QCFractal initialization

        Parameters
        ----------
        name : str, optional
            The name of the server itself, provided when users query information
        port : int, optional
            The port the server will listen on.
        loop : IOLoop, optional
            Provide an IOLoop to use for the server
        compress_response : bool, optional
            Automatic compression of responses, turn on unless behind a proxy that
            provides this capability.
        security : Optional[str], optional
            The security options for the server {None, "local"}. The local security
            option uses the database to cache users.
        allow_read : bool, optional
            Allow unregistered to perform GET operations on Molecule/KeywordSets/KVStore/Results/Procedures
        ssl_options : Optional[Dict[str, str]], optional
            True, automatically creates self-signed SSL certificates. False, turns off SSL entirely. A user can also supply a dictionary of valid certificates.
        storage_uri : str, optional
            The database URI that the underlying storage socket will connect to.
        storage_project_name : str, optional
            The project name to use on the database.
        query_limit : int, optional
            The maximum number of entries a query will return.
        logfile_prefix : str, optional
            The logfile to use for logging.
        loglevel : str, optional
            The level of logging to output
        queue_socket : BaseAdapter, optional
            An optional Adapter to provide for server to have limited local compute.
            Should only be used for testing and interactive sessions.
        heartbeat_frequency : float, optional
            The time (in seconds) of the heartbeat manager frequency.
        max_active_services : int, optional
            The maximum number of active Services that can be running at any given time.
        service_frequency : float, optional
            The time (in seconds) before checking and updating services.
        """

        # Save local options
        self.name = name
        self.port = port
        if ssl_options is False:
            self._address = "http://localhost:" + str(self.port) + "/"
        else:
            self._address = "https://localhost:" + str(self.port) + "/"

        self.max_active_services = max_active_services
        self.service_frequency = service_frequency
        self.heartbeat_frequency = heartbeat_frequency

        # Setup logging.
        if logfile_prefix is not None:
            tornado.options.options["log_file_prefix"] = logfile_prefix

        tornado.log.enable_pretty_logging()
        self.logger = logging.getLogger("tornado.application")
        self.logger.setLevel(loglevel.upper())

        # Create API Access logger class if enables
        if log_apis:
            self.api_logger = API_AccessLogger(geo_file_path=geo_file_path)
        else:
            self.api_logger = None

        # Build security layers
        if security is None:
            storage_bypass_security = True
        elif security == "local":
            storage_bypass_security = False
        else:
            raise KeyError("Security option '{}' not recognized.".format(security))

        # Handle SSL
        ssl_ctx = None
        self.client_verify = True
        if ssl_options is True:
            self.logger.warning("No SSL files passed in, generating self-signed SSL certificate.")
            self.logger.warning("Clients must use `verify=False` when connecting.\n")

            cert, key = _build_ssl()

            # Add quick names
            ssl_name = name.lower().replace(" ", "_")
            cert_name = ssl_name + "_ssl.crt"
            key_name = ssl_name + "_ssl.key"

            ssl_options = {"crt": cert_name, "key": key_name}

            with open(cert_name, "wb") as handle:
                handle.write(cert)

            with open(key_name, "wb") as handle:
                handle.write(key)

            ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ssl_ctx.load_cert_chain(ssl_options["crt"], ssl_options["key"])

            # Destroy keyfiles upon close
            import atexit
            import os

            atexit.register(os.remove, cert_name)
            atexit.register(os.remove, key_name)
            self.client_verify = False

        elif ssl_options is False:
            ssl_ctx = None

        elif isinstance(ssl_options, dict):
            if ("crt" not in ssl_options) or ("key" not in ssl_options):
                raise KeyError("'crt' (SSL Certificate) and 'key' (SSL Key) fields are required for `ssl_options`.")

            ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ssl_ctx.load_cert_chain(ssl_options["crt"], ssl_options["key"])
        else:
            raise KeyError("ssl_options not understood")

        # Setup the database connection
        self.storage_database = storage_project_name
        self.storage_uri = storage_uri
        self.storage = storage_socket_factory(
            storage_uri,
            project_name=storage_project_name,
            bypass_security=storage_bypass_security,
            allow_read=allow_read,
            max_limit=query_limit,
            skip_version_check=skip_storage_version_check,
        )

        if view_enabled:
            self.view_handler = ViewHandler(view_path)
        else:
            self.view_handler = None

        # Pull the current loop if we need it
        self.loop = loop or tornado.ioloop.IOLoop.current()

        # Build up the application
        self.objects = {
            "storage_socket": self.storage,
            "logger": self.logger,
            "api_logger": self.api_logger,
            "view_handler": self.view_handler,
        }

        # Public information
        self.objects["public_information"] = {
            "name": self.name,
            "heartbeat_frequency": self.heartbeat_frequency,
            "version": get_information("version"),
            "query_limit": self.storage.get_limit(1.0e9),
            "client_lower_version_limit": "0.14.0",  # Must be XX.YY.ZZ
            "client_upper_version_limit": "0.15.99",  # Must be XX.YY.ZZ
        }
        self.update_public_information()

        endpoints = [
            # Generic web handlers
            (r"/information", InformationHandler, self.objects),
            (r"/kvstore", KVStoreHandler, self.objects),
            (r"/molecule", MoleculeHandler, self.objects),
            (r"/keyword", KeywordHandler, self.objects),
            (r"/collection(?:/([0-9]+)(?:/(value|entry|list|molecule))?)?", CollectionHandler, self.objects),
            (r"/result", ResultHandler, self.objects),
            (r"/wavefunctionstore", WavefunctionStoreHandler, self.objects),
            (r"/procedure/?", ProcedureHandler, self.objects),
            (r"/optimization/(.*)/?", OptimizationHandler, self.objects),
            # Queue Schedulers
            (r"/task_queue", TaskQueueHandler, self.objects),
            (r"/service_queue", ServiceQueueHandler, self.objects),
            (r"/queue_manager", QueueManagerHandler, self.objects),
            (r"/manager", ComputeManagerHandler, self.objects),
        ]

        # Build the app
        app_settings = {"compress_response": compress_response}
        self.app = tornado.web.Application(endpoints, **app_settings)
        self.endpoints = set([v[0].replace("/", "", 1) for v in endpoints])

        self.http_server = tornado.httpserver.HTTPServer(self.app, ssl_options=ssl_ctx)

        self.http_server.listen(self.port)

        # Add periodic callback holders
        self.periodic = {}

        # Exit callbacks
        self.exit_callbacks = []

        self.logger.info("FractalServer:")
        self.logger.info("    Name:          {}".format(self.name))
        self.logger.info("    Version:       {}".format(get_information("version")))
        self.logger.info("    Address:       {}".format(self._address))
        self.logger.info("    Database URI:  {}".format(storage_uri))
        self.logger.info("    Database Name: {}".format(storage_project_name))
        self.logger.info("    Query Limit:   {}\n".format(self.storage.get_limit(1.0e9)))
        self.loop_active = False

        # Create a executor for background processes
        self.executor = ThreadPoolExecutor(max_workers=2)
        self.futures = {}

        # Queue manager if direct build
        self.queue_socket = queue_socket
        if self.queue_socket is not None:
            if security == "local":
                raise ValueError("Cannot yet use local security with a internal QueueManager")

            def _build_manager():
                client = FractalClient(self, username="qcfractal_server")
                self.objects["queue_manager"] = QueueManager(
                    client,
                    self.queue_socket,
                    logger=self.logger,
                    manager_name="FractalServer",
                    cores_per_task=1,
                    memory_per_task=1,
                    verbose=False,
                )

            # Build the queue manager, will not run until loop starts
            self.futures["queue_manager_future"] = self._run_in_thread(_build_manager)

    def __repr__(self):

        return f"FractalServer(name='{self.name}' uri='{self._address}')"

    def _run_in_thread(self, func, timeout=5):
        """
        Runs a function in a background thread
        """
        if self.executor is None:
            raise AttributeError("No Executor was created, but run_in_thread was called.")

        fut = self.loop.run_in_executor(self.executor, func)
        return fut

    ## Start/stop functionality

    def start(self, start_loop: bool = True, start_periodics: bool = True) -> None:
        """
        Starts up the IOLoop and periodic calls.

        Parameters
        ----------
        start_loop : bool, optional
            If False, does not start the IOLoop
        start_periodics : bool, optional
            If False, does not start the server periodic updates such as
            Service iterations and Manager heartbeat checking.
        """
        if "queue_manager_future" in self.futures:

            def start_manager():
                self._check_manager("manager_build")
                self.objects["queue_manager"].start()

            # Call this after the loop has started
            self._run_in_thread(start_manager)

        # Add services callback
        if start_periodics:
            nanny_services = tornado.ioloop.PeriodicCallback(self.update_services, self.service_frequency * 1000)
            nanny_services.start()
            self.periodic["update_services"] = nanny_services

            # Check Manager heartbeats, 5x heartbeat frequency
            heartbeats = tornado.ioloop.PeriodicCallback(
                self.check_manager_heartbeats, self.heartbeat_frequency * 1000 * 0.2
            )
            heartbeats.start()
            self.periodic["heartbeats"] = heartbeats

            # Log can take some time, update in thread
            def run_log_update_in_thread():
                self._run_in_thread(self.update_server_log)

            server_log = tornado.ioloop.PeriodicCallback(run_log_update_in_thread, self.heartbeat_frequency * 1000)

            server_log.start()
            self.periodic["server_log"] = server_log

        # Build callbacks which are always required
        public_info = tornado.ioloop.PeriodicCallback(self.update_public_information, self.heartbeat_frequency * 1000)
        public_info.start()
        self.periodic["public_info"] = public_info

        # Soft quit with a keyboard interrupt
        self.logger.info("FractalServer successfully started.\n")
        if start_loop:
            self.loop_active = True
            self.loop.start()

    def stop(self, stop_loop: bool = True) -> None:
        """
        Shuts down the IOLoop and periodic updates.

        Parameters
        ----------
        stop_loop : bool, optional
            If False, does not shut down the IOLoop. Useful if the IOLoop is externally managed.
        """

        # Shut down queue manager
        if "queue_manager" in self.objects:
            self._run_in_thread(self.objects["queue_manager"].stop)

        # Close down periodics
        for cb in self.periodic.values():
            cb.stop()

        # Call exit callbacks
        for func, args, kwargs in self.exit_callbacks:
            func(*args, **kwargs)

        # Shutdown executor and futures
        for k, v in self.futures.items():
            v.cancel()

        if self.executor is not None:
            self.executor.shutdown()

        # Shutdown IOLoop if needed
        if (asyncio.get_event_loop().is_running()) and stop_loop:
            self.loop.stop()
        self.loop_active = False

        # Final shutdown
        if stop_loop:
            self.loop.close(all_fds=True)
        self.logger.info("FractalServer stopping gracefully. Stopped IOLoop.\n")

    def add_exit_callback(self, callback, *args, **kwargs):
        """Adds additional callbacks to perform when closing down the server.

        Parameters
        ----------
        callback : callable
            The function to call at exit
        *args
            Arguments to call with the function.
        **kwargs
            Kwargs to call with the function.

        """
        self.exit_callbacks.append((callback, args, kwargs))

    ## Helpers

    def get_address(self, endpoint: Optional[str] = None) -> str:
        """Obtains the full URI for a given function on the FractalServer.

        Parameters
        ----------
        endpoint : Optional[str], optional
            Specifies a endpoint to provide the URI for. If None returns the server address.

        Returns
        -------
        str
            The endpoint URI

        """

        if endpoint and (endpoint not in self.endpoints):
            raise AttributeError("Endpoint '{}' not found.".format(endpoint))

        if endpoint:
            return self._address + endpoint
        else:
            return self._address

    ## Updates

    def update_services(self) -> int:
        """Runs through all active services and examines their current status."""

        # Grab current services
        current_services = self.storage.get_services(status="RUNNING")["data"]

        # Grab new services if we have open slots
        open_slots = max(0, self.max_active_services - len(current_services))
        if open_slots > 0:
            new_services = self.storage.get_services(status="WAITING", limit=open_slots)["data"]
            current_services.extend(new_services)
            if len(new_services):
                self.logger.info(f"Starting {len(new_services)} new services.")

        self.logger.debug(f"Updating {len(current_services)} services.")

        # Loop over the services and iterate
        running_services = 0
        completed_services = []
        for data in current_services:

            # TODO HACK: remove task_id from 'output'. This is contained in services
            # created in previous versions. Doing this now, but should do a db migration
            # at some point
            if "output" in data:
                data["output"].pop("task_id", None)

            # Attempt to iteration and get message
            try:
                service = construct_service(self.storage, self.logger, data)
                finished = service.iterate()
            except Exception:
                error_message = "FractalServer Service Build and Iterate Error:\n{}".format(traceback.format_exc())
                self.logger.error(error_message)
                service.status = "ERROR"
                service.error = ComputeError(error_type="iteration_error", error_message=error_message)
                finished = False

            self.storage.update_services([service])

            # Mark procedure and service as error
            if service.status == "ERROR":
                self.storage.update_service_status("ERROR", id=service.id)

            if finished is not False:
                # Add results to procedures, remove complete_ids
                completed_services.append(service)
            else:
                running_services += 1

        if len(completed_services):
            self.logger.info(f"Completed {len(completed_services)} services.")

        self.logger.debug(f"Done updating services.")

        # Add new procedures and services
        self.storage.services_completed(completed_services)

        return running_services

    def update_server_log(self) -> Dict[str, Any]:
        """
        Updates the servers internal log
        """

        return self.storage.log_server_stats()

    def update_public_information(self) -> None:
        """
        Updates the public information data
        """
        data = self.storage.get_server_stats_log(limit=1)["data"]

        counts = {"collection": 0, "molecule": 0, "result": 0, "kvstore": 0}
        if len(data):
            counts["collection"] = data[0].get("collection_count", 0)
            counts["molecule"] = data[0].get("molecule_count", 0)
            counts["result"] = data[0].get("result_count", 0)
            counts["kvstore"] = data[0].get("kvstore_count", 0)

        update = {"counts": counts}
        self.objects["public_information"].update(update)

    def check_manager_heartbeats(self) -> None:
        """
        Checks the heartbeats and kills off managers that have not been heard from.
        """

        dt = datetime.datetime.utcnow() - datetime.timedelta(seconds=self.heartbeat_frequency)
        ret = self.storage.get_managers(status="ACTIVE", modified_before=dt)

        for blob in ret["data"]:
            nshutdown = self.storage.queue_reset_status(manager=blob["name"], reset_running=True)
            self.storage.manager_update(blob["name"], returned=nshutdown, status="INACTIVE")

            self.logger.info(
                "Hearbeat missing from {}. Shutting down, recycling {} incomplete tasks.".format(
                    blob["name"], nshutdown
                )
            )

    def list_managers(self, status: Optional[str] = None, name: Optional[str] = None) -> List[Dict[str, Any]]:
        """
        Provides a list of managers associated with the server both active and inactive.

        Parameters
        ----------
        status : Optional[str], optional
            Filters managers by status.
        name : Optional[str], optional
            Filters managers by name

        Returns
        -------
        List[Dict[str, Any]]
            The requested Manager data.
        """

        return self.storage.get_managers(status=status, name=name)["data"]

    def client(self):
        """
        Builds a client from this server.
        """

        return FractalClient(self)

    ### Functions only available if using a local queue_adapter

    def _check_manager(self, func_name: str) -> None:
        if self.queue_socket is None:
            raise AttributeError(
                "{} is only available if the server was initialized with a queue manager.".format(func_name)
            )

        # Wait up to one second for the queue manager to build
        if "queue_manager" not in self.objects:
            self.logger.info("Waiting on queue_manager to build.")
            for x in range(20):
                time.sleep(0.1)
                if "queue_manager" in self.objects:
                    break

            if "queue_manager" not in self.objects:
                raise AttributeError("QueueManager never constructed.")

    def update_tasks(self) -> bool:
        """Pulls tasks from the queue_adapter, inserts them into the database,
        and fills the queue_adapter with new tasks.

        Returns
        -------
        bool
            Return True if the operation completed successfully
        """
        self._check_manager("update_tasks")

        if self.loop_active:
            # Drop this in a thread so that we are not blocking each other
            self._run_in_thread(self.objects["queue_manager"].update)
        else:
            self.objects["queue_manager"].update()

        return True

    def await_results(self) -> bool:
        """A synchronous method for testing or small launches
        that awaits task completion before adding all queued results
        to the database and returning.

        Returns
        -------
        bool
            Return True if the operation completed successfully
        """
        self._check_manager("await_results")

        self.logger.info("Updating tasks")
        return self.objects["queue_manager"].await_results()

    def await_services(self, max_iter: int = 10) -> bool:
        """A synchronous method that awaits the completion of all services
        before returning.

        Parameters
        ----------
        max_iter : int, optional
            The maximum number of service iterations the server will run through. Will
            terminate early if all services have completed.

        Returns
        -------
        bool
            Return True if the operation completed successfully

        """
        self._check_manager("await_services")

        self.await_results()
        for x in range(1, max_iter + 1):
            self.logger.info("\nAwait services: Iteration {}\n".format(x))
            running_services = self.update_services()
            self.await_results()
            if running_services == 0:
                break

        return True

    def list_current_tasks(self) -> List[Any]:
        """Provides a list of tasks currently in the queue along
        with the associated keys.

        Returns
        -------
        ret : list of tuples
            All tasks currently still in the database
        """
        self._check_manager("list_current_tasks")

        return self.objects["queue_manager"].list_current_tasks()



"""
Utility functions for handling network ports
"""

import socket


def find_port() -> int:
    sock = socket.socket()
    sock.bind(("localhost", 0))
    host, port = sock.getsockname()
    return port


def is_port_open(ip: str, port: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.settimeout(1)
        try:
            s.connect((ip, int(port)))
            s.shutdown(socket.SHUT_RDWR)
            return True
        except ConnectionRefusedError:
            return False



"""
Tests the on-node procedures compute capabilities.
"""

import copy

import pytest

import qcfractal.interface as ptl
from qcfractal.interface.models import GridOptimizationInput, TorsionDriveInput
from qcfractal.testing import fractal_compute_server, recursive_dict_merge, using_geometric, using_rdkit


@pytest.fixture(scope="module")
def torsiondrive_fixture(fractal_compute_server):

    # Cannot use this fixture without these services. Also cannot use `mark` and `fixture` decorators
    pytest.importorskip("torsiondrive")
    pytest.importorskip("geometric")
    pytest.importorskip("rdkit")

    client = ptl.FractalClient(fractal_compute_server)

    # Add a HOOH
    hooh = ptl.data.get_molecule("hooh.json")
    mol_ret = client.add_molecules([hooh])

    # Geometric options
    torsiondrive_options = {
        "initial_molecule": mol_ret[0],
        "keywords": {"dihedrals": [[0, 1, 2, 3]], "grid_spacing": [90]},
        "optimization_spec": {
            "program": "geometric",
            "keywords": {"coordsys": "tric"},
            "protocols": {"trajectory": "initial_and_final"},
        },
        "qc_spec": {"driver": "gradient", "method": "UFF", "basis": "", "keywords": None, "program": "rdkit"},
    }  # yapf: disable

    def spin_up_test(**keyword_augments):
        run_service = keyword_augments.pop("run_service", True)

        instance_options = copy.deepcopy(torsiondrive_options)
        recursive_dict_merge(instance_options, keyword_augments)

        inp = TorsionDriveInput(**instance_options)
        ret = client.add_service([inp], full_return=True)

        if ret.meta.n_inserted:  # In case test already submitted
            compute_key = ret.data.ids[0]
            service = client.query_services(procedure_id=compute_key)[0]
            assert "WAITING" in service["status"]

        if run_service:
            fractal_compute_server.await_services()
            assert len(fractal_compute_server.list_current_tasks()) == 0

        return ret.data

    yield spin_up_test, client


def test_service_torsiondrive_service_incomplete(fractal_compute_server, torsiondrive_fixture):
    hooh = ptl.data.get_molecule("hooh.json")
    hooh.geometry[0] += 0.00031

    spin_up_test, client = torsiondrive_fixture
    ret = spin_up_test(run_service=False)

    # Check the blank
    result = client.query_procedures(id=ret.ids)[0]
    assert len(result.final_energy_dict) == 0
    assert len(result.optimization_history) == 0
    assert result.status == "INCOMPLETE"

    # Update the service, but no compute
    fractal_compute_server.update_services()
    result = client.query_procedures(id=ret.ids)[0]
    status = result.detailed_status()
    assert result.status == "RUNNING"
    assert status["incomplete_tasks"] == 1

    fractal_compute_server.await_results()

    # Take a compute step
    fractal_compute_server.await_services(max_iter=1)
    result = client.query_procedures(id=ret.ids)[0]
    status = result.detailed_status()
    assert status["total_points"] == 4
    assert status["computed_points"] == 3
    assert status["complete_tasks"] >= 3
    assert status["incomplete_tasks"] == 0
    assert len(result.final_energy_dict) == 1  # One complete
    assert len(result.optimization_history) == 3  # Three spawned
    assert result.minimum_positions["[-90]"] == 0
    assert result.status == "RUNNING"

    # Repeat compute step checking for updates
    fractal_compute_server.await_services(max_iter=1)
    result = client.query_procedures(id=ret.ids)[0]
    assert len(result.final_energy_dict) == 3
    assert len(result.optimization_history) == 4
    assert result.minimum_positions["[-90]"] == 0
    assert result.status == "RUNNING"

    # Finalize
    fractal_compute_server.await_services(max_iter=6)
    result = client.query_procedures(id=ret.ids)[0]
    assert len(result.final_energy_dict) == 4
    assert len(result.optimization_history) == 4
    assert result.minimum_positions["[-90]"] == 2
    assert result.status == "COMPLETE"


def test_service_manipulation(torsiondrive_fixture):

    spin_up_test, client = torsiondrive_fixture

    hooh = ptl.data.get_molecule("hooh.json")
    hooh.geometry[0] += 3.1

    ret = spin_up_test(run_service=False, initial_molecule=[hooh])

    service = client.query_services(procedure_id=ret.ids)[0]
    assert service["status"] == "WAITING"

    client.modify_services("restart", id=service["id"])

    service = client.query_services(procedure_id=ret.ids)[0]
    assert service["status"] == "RUNNING"


def test_service_torsiondrive_single(torsiondrive_fixture):
    """"Tests torsiondrive pathway and checks the result """

    spin_up_test, client = torsiondrive_fixture

    ret = spin_up_test()

    # Get a TorsionDriveORM result and check data
    result = client.query_procedures(id=ret.ids)[0]
    assert result.status == "COMPLETE"
    assert isinstance(str(result), str)  # Check that repr runs

    assert pytest.approx(0.002597541340221565, abs=1e-6) == result.get_final_energies(0)
    assert pytest.approx(0.000156553761859276, abs=1e-6) == result.get_final_energies(90)
    assert pytest.approx(0.000156553761859271, abs=1e-6) == result.get_final_energies(-90)
    assert pytest.approx(0.000753492556057886, abs=1e-6) == result.get_final_energies(180)

    assert hasattr(result.get_final_molecules()[(-90,)], "symbols")

    opt = result.get_history(0)[0]
    assert opt.protocols.trajectory == "initial_and_final"
    assert len(opt.trajectory) == 2


@pytest.mark.slow
def test_service_torsiondrive_multi_single(torsiondrive_fixture):
    spin_up_test, client = torsiondrive_fixture

    hooh = ptl.data.get_molecule("hooh.json")
    hooh2 = hooh.copy(deep=True)
    hooh2.geometry[0] += 0.0004

    ret = spin_up_test(initial_molecule=[hooh, hooh2])

    result = client.query_procedures(id=ret.ids)[0]
    assert result.status == "COMPLETE"


def test_service_torsiondrive_duplicates(torsiondrive_fixture):
    """Ensure that duplicates are properly caught and yield the same results without calculation"""

    spin_up_test, client = torsiondrive_fixture

    # Run the test without modifications
    id1 = spin_up_test().ids[0]

    # Augment the input for torsion drive to yield a new hash procedure hash,
    # but not a new task set
    id2 = spin_up_test(keywords={"energy_upper_limit": 1000}).ids[0]

    assert id1 != id2
    procedures = client.query_procedures(id=[id1, id2])
    assert len(procedures) == 2  # Make sure only 2 procedures are yielded

    base_run, duplicate_run = procedures
    assert base_run.optimization_history == duplicate_run.optimization_history


@pytest.mark.slow
def test_service_torsiondrive_option_dihedral_ranges(torsiondrive_fixture):
    """"Tests torsiondrive with dihedral_ranges optional keyword """

    spin_up_test, client = torsiondrive_fixture
    ret = spin_up_test(keywords={"grid_spacing": [30], "dihedral_ranges": [[-150, -60]]})

    result = client.query_procedures(id=ret.ids)[0]
    assert result.status == "COMPLETE"

    # The dihedral range should be limited to -150, -90, -60
    final_energies = result.get_final_energies()
    assert set(final_energies.keys()) == {(-150,), (-120,), (-90,), (-60,)}
    assert pytest.approx(0.0005683235570009067, abs=1e-6) == final_energies[(-150,)]
    assert pytest.approx(0.0002170694130912583, abs=1e-6) == final_energies[(-120,)]
    assert pytest.approx(0.0001565537585121726, abs=1e-6) == final_energies[(-90,)]
    assert pytest.approx(0.0007991274441437338, abs=1e-6) == final_energies[(-60,)]

    # Check final molecules
    final_molecules = result.get_final_molecules()
    assert set(final_molecules.keys()) == {(-150,), (-120,), (-90,), (-60,)}
    assert all(hasattr(m, "symbols") for m in final_molecules.values())


@pytest.mark.slow
def test_service_torsiondrive_option_energy_decrease_thresh(torsiondrive_fixture):
    """"Tests torsiondrive with energy_decrease_thresh optional keyword"""

    spin_up_test, client = torsiondrive_fixture
    ret = spin_up_test(keywords={"grid_spacing": [90], "energy_decrease_thresh": 3e-5})

    result = client.query_procedures(id=ret.ids)[0]
    assert result.status == "COMPLETE"

    # the final energies are the same as the default setting, because this molecule is too simple
    final_energies = result.get_final_energies()
    assert set(final_energies.keys()) == {(-90,), (-0,), (90,), (180,)}
    assert pytest.approx(0.002597541340221565, abs=1e-6) == final_energies[(0,)]
    assert pytest.approx(0.000156553761859276, abs=1e-6) == final_energies[(90,)]
    assert pytest.approx(0.000156553761859271, abs=1e-6) == final_energies[(-90,)]
    assert pytest.approx(0.000753492556057886, abs=1e-6) == final_energies[(180,)]


@pytest.mark.slow
def test_service_torsiondrive_option_energy_upper_limit(torsiondrive_fixture):
    """"Tests torsiondrive with energy_upper_limit optional keyword"""

    spin_up_test, client = torsiondrive_fixture
    ret = spin_up_test(keywords={"grid_spacing": [30], "energy_upper_limit": 1e-4})

    result = client.query_procedures(id=ret.ids)[0]
    assert result.status == "COMPLETE"

    # The energy_upper_limit should limit the range of the scan
    final_energies = result.get_final_energies()
    assert set(final_energies.keys()) == {(-150,), (-120,), (-90,), (-60,)}
    assert pytest.approx(0.0005683235570009067, abs=1e-6) == final_energies[(-150,)]
    assert pytest.approx(0.0002170694130912583, abs=1e-6) == final_energies[(-120,)]
    assert pytest.approx(0.0001565537585121726, abs=1e-6) == final_energies[(-90,)]
    assert pytest.approx(0.0007991274441437338, abs=1e-6) == final_energies[(-60,)]


@pytest.mark.slow
def test_service_torsiondrive_option_extra_constraints(torsiondrive_fixture):
    """"Tests torsiondrive with extra_constraints in optimization_spec """

    spin_up_test, client = torsiondrive_fixture
    ret = spin_up_test(
        optimization_spec={
            "program": "geometric",
            "keywords": {"coordsys": "tric", "constraints": {"freeze": [{"type": "xyz", "indices": [0]}]}},
        }
    )

    result = client.query_procedures(id=ret.ids)[0]
    assert result.status == "COMPLETE"

    # The final energies are the same as the default setting, because this molecule is too simple
    final_energies = result.get_final_energies()
    assert set(final_energies.keys()) == {(-90,), (-0,), (90,), (180,)}
    assert pytest.approx(0.002597541340221565, abs=1e-6) == final_energies[(0,)]
    assert pytest.approx(0.000156553761859276, abs=1e-6) == final_energies[(90,)]
    assert pytest.approx(0.000156553761859271, abs=1e-6) == final_energies[(-90,)]
    assert pytest.approx(0.000753492556057886, abs=1e-6) == final_energies[(180,)]

    # Check final molecules
    hooh = ptl.data.get_molecule("hooh.json")
    final_molecules = result.get_final_molecules()
    for m in final_molecules.values():
        # the coordinate of the first atom should be "frozen"
        assert pytest.approx(m.geometry[0], abs=1e-3) == hooh.geometry[0]


def test_service_iterate_error(torsiondrive_fixture):
    """Ensure errors are caught and logged when iterating serivces"""

    spin_up_test, client = torsiondrive_fixture

    # Run the test without modifications
    ret = spin_up_test(keywords={"dihedrals": [[0, 1, 2, 50]]})

    status = client.query_services(procedure_id=ret.ids)
    assert len(status) == 1

    assert status[0]["status"] == "ERROR"
    assert "Service Build" in status[0]["error"]["error_message"]

    # Test that the error is propagated to the procedure
    proc_status = client.query_procedures(ret.ids)
    assert len(proc_status) == 1
    assert proc_status[0].status == "ERROR"
    assert "Service Build" in proc_status[0].get_error().error_message


def test_service_torsiondrive_compute_error(torsiondrive_fixture):
    """Ensure errors are caught and logged when computing serivces"""

    spin_up_test, client = torsiondrive_fixture

    # Run the test without modifications
    ret = spin_up_test(qc_spec={"method": "waffles_crasher"})

    status = client.query_services(procedure_id=ret.ids)
    assert len(status) == 1

    assert status[0]["status"] == "ERROR"
    assert "All tasks" in status[0]["error"]["error_message"]


def test_service_torsiondrive_visualization(torsiondrive_fixture):
    """Test the visualization function for the 1-D case"""

    spin_up_test, client = torsiondrive_fixture
    ret = spin_up_test()

    # Get a TorsionDriveORM result and check data
    result = client.query_procedures(id=ret.ids)[0]
    assert result.status == "COMPLETE"

    result.visualize()


def test_service_torsiondrive_get_final_results(torsiondrive_fixture):
    """Test the get_final_results function for the 1-D case"""

    spin_up_test, client = torsiondrive_fixture
    ret = spin_up_test()

    # Get a TorsionDriveORM result and check data
    result = client.query_procedures(id=ret.ids)[0]
    assert result.status == "COMPLETE"

    final_result_records = result.get_final_results()
    assert set(final_result_records.keys()) == {(-90,), (-0,), (90,), (180,)}

    stdout = result.get_stdout()
    assert "All optimizations converged at lowest energy" in stdout


@using_geometric
@using_rdkit
def test_service_gridoptimization_single_opt(fractal_compute_server):

    client = ptl.FractalClient(fractal_compute_server)

    # Add a HOOH
    hooh = ptl.data.get_molecule("hooh.json")
    initial_distance = hooh.measure([1, 2])
    mol_ret = client.add_molecules([hooh])

    # Options
    service = GridOptimizationInput(
        **{
            "keywords": {
                "preoptimization": True,
                "scans": [
                    {"type": "distance", "indices": [1, 2], "steps": [-0.1, 0.0], "step_type": "relative"},
                    {"type": "dihedral", "indices": [0, 1, 2, 3], "steps": [-90, 0], "step_type": "absolute"},
                ],
            },
            "optimization_spec": {"program": "geometric", "keywords": {"coordsys": "tric"}},
            "qc_spec": {"driver": "gradient", "method": "UFF", "basis": "", "keywords": None, "program": "rdkit"},
            "initial_molecule": mol_ret[0],
        }
    )  # yapf: disable

    ret = client.add_service([service], tag="gridopt", priority="low")

    fractal_compute_server.await_services(max_iter=1)
    result = client.query_procedures(id=ret.ids)[0]
    assert result.grid_optimizations.keys() == {'"preoptimization"'}
    assert result.status == "RUNNING"

    fractal_compute_server.await_services(max_iter=1)
    result = client.query_procedures(id=ret.ids)[0]
    status = result.detailed_status()
    assert status["total_points"] == 5
    assert status["complete_tasks"] == 2
    assert result.grid_optimizations.keys() == {'"preoptimization"', "[1, 0]"}
    assert result.status == "RUNNING"

    fractal_compute_server.await_services(max_iter=1)
    result = client.query_procedures(id=ret.ids)[0]
    assert result.grid_optimizations.keys() == {'"preoptimization"', "[1, 0]", "[0, 0]", "[1, 1]"}
    assert result.status == "RUNNING"

    fractal_compute_server.await_services(max_iter=6)
    result = client.query_procedures(id=ret.ids)[0]
    status = result.detailed_status()
    assert status["complete_tasks"] == 5
    assert result.status == "COMPLETE"

    assert result.starting_grid == (1, 0)
    assert pytest.approx(result.get_final_energies((0, 0)), abs=1.0e-4) == 0.0010044105443485617
    assert pytest.approx(result.get_final_energies((1, 1)), abs=1.0e-4) == 0.0026440964897817623

    assert result.starting_molecule != result.initial_molecule

    # Check initial vs starting molecule
    assert result.initial_molecule == mol_ret[0]
    starting_mol = client.query_molecules(id=result.starting_molecule)[0]
    assert pytest.approx(starting_mol.measure([1, 2])) != initial_distance
    assert pytest.approx(starting_mol.measure([1, 2])) == 2.488686479260597

    # Check tags on individual procedures
    proc_id = result.grid_optimizations["[0, 0]"]

    # completed tasks should be deleted
    task = client.query_tasks(base_result=proc_id)
    assert not task

    # Check final ResultRecords
    final_result_records = result.get_final_results()
    assert len(final_result_records) == 5
    assert final_result_records["preoptimization"].molecule == result.starting_molecule

    # Pull the full history
    preopt = result.get_history()["preoptimization"]
    assert preopt.initial_molecule == result.initial_molecule
    assert preopt.final_molecule == result.starting_molecule


@using_geometric
@using_rdkit
def test_service_gridoptimization_single_noopt(fractal_compute_server):

    client = ptl.FractalClient(fractal_compute_server)

    # Add a HOOH
    hooh = ptl.data.get_molecule("hooh.json")
    initial_distance = hooh.measure([1, 2])

    # Options
    service = GridOptimizationInput(
        **{
            "keywords": {
                "preoptimization": False,
                "scans": [{"type": "distance", "indices": [1, 2], "steps": [-0.1, 0.0], "step_type": "relative"}],
            },
            "optimization_spec": {"program": "geometric", "keywords": {"coordsys": "tric"}},
            "qc_spec": {"driver": "gradient", "method": "UFF", "basis": "", "keywords": None, "program": "rdkit"},
            "initial_molecule": hooh,
        }
    )  # yapf: disable

    ret = client.add_service([service])
    fractal_compute_server.await_services()
    assert len(fractal_compute_server.list_current_tasks()) == 0

    result = client.query_procedures(id=ret.ids)[0]

    assert result.status == "COMPLETE"
    assert result.starting_grid == (1,)
    assert pytest.approx(result.get_final_energies((0,)), abs=1.0e-4) == 0.00032145876568280524

    assert result.starting_molecule == result.initial_molecule

    # Check initial vs startin molecule
    assert result.initial_molecule == result.starting_molecule

    mol = client.query_molecules(id=result.starting_molecule)[0]
    assert pytest.approx(mol.measure([1, 2])) == initial_distance


@pytest.mark.skip
def test_query_time(fractal_compute_server):

    client = ptl.FractalClient(fractal_compute_server)

    import time

    t = time.time()
    p = client.query_procedures(procedure="optimization")
    total = time.time() - t
    print(total)
    print(len(p))
    print("---- per optmization proc: ", total / len(p))

    import time

    t = time.time()
    p = client.query_procedures(procedure="torsiondrive")
    total = time.time() - t
    print(total)
    print(len(p))
    print("---- per Torsion proc: ", total / len(p))
